from re import compile

from eis1600.helper.EntityTags import EntityTags

AR_LETTERS_CHARSET = frozenset(
        u'\u0621\u0622\u0623\u0624\u0625\u0626\u0627'
        u'\u0628\u0629\u062a\u062b\u062c\u062d\u062e'
        u'\u062f\u0630\u0631\u0632\u0633\u0634\u0635'
        u'\u0636\u0637\u0638\u0639\u063a\u0640\u0641'
        u'\u0642\u0643\u0644\u0645\u0646\u0647\u0648'
        u'\u0649\u064a\u0671\u067e\u0686\u06a4\u06af'
)
AR_CHR = r'[' + u''.join(AR_LETTERS_CHARSET) + ']'
AR_STR = AR_CHR + '+'
AR_STR_AND_TAGS = r'[' + u''.join(AR_LETTERS_CHARSET) + 'a-zA-ZÜ0-9]+'
WORD = r'(?:(^|\s)' + AR_STR + ')'
NOISE_ELEMENTS = r'(?: [\[\]0-9،.():~|-])*'

# EIS1600 mARkdown
UID = r'_ء_(#)?=(?P<UID>\d{12})= '
UID_PATTERN = compile(UID)
MIU_UID = r'_ء_#=(?P<UID>\d{12})= '
MIU_UID_PATTERN = compile(MIU_UID)
HEADER_END_PATTERN = compile(r'(#META#Header#End#)\n')
MIU_HEADER = r'#MIU#Header#'
MIU_HEADER_PATTERN = compile(MIU_HEADER)
HEADING_PATTERN = compile(UID + r'(?P<level>[|]+) (?P<heading>.*)\n')
EMPTY_PARAGRAPH_PATTERN = compile(UID + r'::[A-Z]+:: ~')
EMPTY_FIRST_PARAGRAPH_PATTERN = compile(r'_ء_#=\d{12}=')
PAGE_TAG = r' ?(?P<page_tag>PageV\d{2}P\d{3,})'
PAGE_TAG_PATTERN = compile(PAGE_TAG)
ONLY_PAGE_TAG = UID + r'::[A-Z]+:: ~\n' + PAGE_TAG
ONLY_PAGE_TAG_PATTERN = compile(ONLY_PAGE_TAG)
PAGE_TAG_IN_BETWEEN_PATTERN = compile(
        AR_STR + r' ?' + r'\n\n' + ONLY_PAGE_TAG + r'\n\n' + r'_ء_=\d{12}= ::[A-Z]+:: ~\n' + AR_STR
)
TEXT_STARTS_WITH_PARAGRAPH = compile(r'_ء_=')
PARAGRAPH_TAG_MISSING = compile('(\n\n[^_])|(\n\n' + MIU_UID + '[^\n]+\n(?:_ء_ )?)' + AR_CHR)
SIMPLE_MARKDOWN = compile('\n#')
SPAN_ELEMENTS = compile('<\\*span>')
POETRY_ATTACHED_AFTER_PAGE_TAG = compile('Page[VP0-9]+[^\n%]+%')

# MIU_TAG_PATTERN is used to split text - indices depend on the number of capturing groups so be careful when
# changing them
MIU_TAG_PATTERN = compile(r'(' + MIU_UID + r'(?P<category>[^\n]+))')
CATEGORY_PATTERN = compile(r'[$|@]+(?:[A-Z_]+[|$])?')
SECTION_TAG = r'_ء_=\d{12}= ::[A-Z]+:: ~'
SECTION_PATTERN = compile(SECTION_TAG)
SECTION_SPLITTER_PATTERN = compile(r'\n\n(' + SECTION_TAG + ')\n(?:_ء_)?')
TAG_PATTERN = compile(r'Ü?(?:[a-zA-Z_%~]+(?:\.[a-zA-Z0-9_%~]+)?)|' + PAGE_TAG + '|(?:::)')
NOR_DIGIT_NOR_AR_STR = r'[^\d\n' + u''.join(AR_LETTERS_CHARSET) + ']+?'
TAG_AND_TEXT_SAME_LINE = r'([$@]+' + NOR_DIGIT_NOR_AR_STR + r'\d*' + NOR_DIGIT_NOR_AR_STR + r') ?((?:[(\[] ?)?' + AR_STR + r')'
UID_TAG_AND_TEXT_SAME_LINE_PATTERN = compile(
        r'(_ء_#=\d{12}= )' + TAG_AND_TEXT_SAME_LINE)

# Catches MIU tags for BIO, CHR and PARATEXT, EDITOR, etc. (everything in between pipes).
# Does not catch HEADERS!
MIU_TAG_AND_TEXT_PATTERN = compile(r'(' + MIU_UID + r'(?:[$@]+?|\|[A-Z]+\|)(?: \d+)?)\n((?:\( ?)?' + AR_STR + r')')

# MIU entity tags
entity_tags = '|'.join(EntityTags().get_entity_tags())
ENTITY_TAGS_PATTERN = compile(r'\bÜ?(?P<full_tag>'
                              r'(?P<entity>' + entity_tags + r')(?P<length>\d{1,2})'
                                                             r'(?:(?P<sub_cat>[A-Z]+)|['r'A-Z0-9]+)?)\b')
YEAR_PATTERN = compile(r'Ü?Y(?P<num_tokens>\d{1,2})(?P<cat>[A-Z])(?P<written>\d{4}|None)(?P<i>I)?Y(?P<real>\d{4})?')
AGE_PATTERN = compile(r'Ü?A\d(?P<cat>[A-Z])(?P<written>\d{2,3})(?P<i>I)?A(?P<real>\d{2,3})?')
TOPONYM_PATTERN = compile(r'Ü?T(?P<num_tokens>\d{1,2})(?P<cat>[A-Z])')
onom_tags = '|'.join(EntityTags().get_onom_tags())
ONOM_TAGS_PATTERN = compile(r'Ü?(?P<entity>' + onom_tags + r')(?P<length>\d{1,2})')

# EIS1600 light mARkdown
HEADING_OR_BIO_PATTERN = compile(r'# [|$]+')
MIU_LIGHT_OR_EIS1600_PATTERN = compile(r'#|_ء_#')
PAGE_TAG_ON_NEWLINE_TMP_PATTERN = compile(r'(?<!\n)\n' + PAGE_TAG + r'(?=\n)')

# Fix mARkdown files
SPACES_CROWD_PATTERN = compile(r'  +')
NEWLINES_CROWD_PATTERN = compile(r'\n{3,}')
NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN = compile(r'[^\n]\n(?:(?:# [|$])|(?:' + UID + '))')
MISSING_DIRECTION_TAG_PATTERN = compile(r'(\n+)(' + AR_CHR + ')')
SPACES_AFTER_NEWLINES_PATTERN = compile(r'\n +')
POETRY_PATTERN = compile(
        r'# (' + AR_STR_AND_TAGS + '(?: ' + AR_STR_AND_TAGS + ')* %~% ' + AR_STR_AND_TAGS + '(?: ' +
        AR_STR_AND_TAGS +
        r')*) ?'
)
BELONGS_TO_PREV_PARAGRAPH_PATTERN = compile(r'\n(.{1,10})\n')
PAGE_TAG_ON_NEWLINE_MARKDOWN_PATTERN = compile(r'\n' + PAGE_TAG)
PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN = compile(
        '(' + AR_STR + ' ?)' + r'\n\n' + PAGE_TAG + r'\n\n' + '(' + AR_STR +
        ')'
)
NORMALIZE_BIO_CHR_MD_PATTERN = compile('# ([$@]((BIO|CHR)_[A-Z]+[$@])| RAW)')
BIO_CHR_TO_NEWLINE_PATTERN = compile(TAG_AND_TEXT_SAME_LINE)

# Fixed poetry old file path pattern
FIXED_POETRY_OLD_PATH_PATTERN = compile(r'/Users/romanov/_OpenITI/_main_corpus/\w+/data/')

print()


from functools import partial
from logging import ERROR, Formatter
from glob import glob

from eis1600.helper.logging import setup_logger

from eis1600.miu.methods import disassemble_text

from eis1600.markdown.methods import insert_uids
from p_tqdm import p_uimap

from pandas import Series, read_csv

from eis1600.helper.repo import MIU_REPO, TEXT_REPO


def main():
    filepath = TEXT_REPO + '_EIS1600 - Text Selection - Serial Source Test - EIS1600_AutomaticSelectionForReview.csv'

    df = read_csv(filepath, usecols=['Book Title', 'PREPARED']).dropna()
    df_ready = df.loc[df['PREPARED'].str.fullmatch('ready')]
    df_double_checked = df.loc[df['PREPARED'].str.fullmatch('double-checked')]

    print(len(df_ready))
    print(len(df_double_checked))

    double_checked_files = []
    ready_files = []

    print('URIs for double-checked files for whom no .EIS1600 file was found')
    for uri in df_double_checked['Book Title']:
        author, text = uri.split('.')
        text_path = TEXT_REPO + 'data/' + author + '/' + uri + '/'
        text_file = glob(text_path + '*.EIS1600')
        if text_file:
            double_checked_files.append(text_file[0])
        else:
            print(uri)

    print('URIs for ready files for whom no .EIS1600TMP file was found')
    for uri in df_ready['Book Title']:
        author, text = uri.split('.')
        text_path = TEXT_REPO + 'data/' + author + '/' + uri + '/'
        tmp_file = glob(text_path + '*.EIS1600TMP')
        eis_file = glob(text_path + '*.EIS1600')
        if tmp_file and not eis_file:
            ready_files.append(tmp_file[0])
        elif tmp_file and eis_file:
            double_checked_files.append(eis_file[0])
            # print(f'{uri} (both TMP and EIS1600)')
        elif eis_file and not tmp_file:
            double_checked_files.append(eis_file[0])
            print(f'{uri} (no TMP but EIS1600)')
        else:
            print(f'{uri} (missing)')

    print('Insert_UIDs into ready texts')

    x = 0
    for i, file in enumerate(ready_files[x:]):
        print(i + x, file)
        insert_uids(file)

    # res = []
    # res += p_uimap(insert_uids, ready_files)

    print('Disassemble double-checked and ready texts')

    texts = double_checked_files + [r.replace('TMP', '') for r in ready_files]
    out_path = MIU_REPO + 'data/'

    x = 0
    # 110 is not cleaned yet

    formatter = Formatter('%(message)s\n\n\n')
    logger = setup_logger('mal_formatted_texts', TEXT_REPO + 'mal_formatted_texts.log', ERROR, formatter)
    count = 0
    csv = []
    for i, text in enumerate(texts[x:]):
        print(i + x, text)
        try:
            disassemble_text(text, out_path)
        except ValueError as e:
            csv.append(text)
            count += 1
            logger.error(e)

    series = Series(csv, name='mal-formatted text')
    series.to_csv(TEXT_REPO + 'mal-formatted-texts.csv', index=False)

    print(f'{count} texts need fixing')

    # res = []
    # res += p_uimap(partial(disassemble_text, out_path=out_path), texts)

    print('Done')




from logging import ERROR
from sys import argv, exit
from os.path import isfile, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from functools import partial

from eis1600.helper.logging import setup_logger

from p_tqdm import p_uimap
from tqdm import tqdm

from eis1600.helper.repo import MIU_REPO, get_path_to_other_repo, read_files_from_autoreport, \
    get_files_from_eis1600_dir, \
    write_to_readme, TEXT_REPO
from eis1600.miu.methods import disassemble_text


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.EIS1600':
                parser.error('You need to input an EIS1600 file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to disassemble EIS1600 file(s) into MIU file(s).
-----
Give a single EIS1600 file as input
or 
Run without input arg to batch process all double-checked EIS1600 files from the AUTOREPORT.
'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    debug = args.debug
    errors = False

    if args.input:
        infile = './' + args.input
        out_path = get_path_to_other_repo(infile, 'MIU')
        print(f'Disassemble {infile}')
        disassemble_text(infile, out_path, debug)
        infiles = [infile.split('/')[-1]]
        path = out_path.split('data')[0]
        write_to_readme(path, infiles, '# Texts disassembled into MIU files\n')
    else:
        input_dir = TEXT_REPO
        out_path = MIU_REPO + 'data/'

        print(f'Disassemble double-checked EIS1600 files from the AUTOREPORT')
        files_list = read_files_from_autoreport(input_dir)

        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600')
        if not infiles:
            print('There are no EIS1600 files to process')
            exit()

        if debug:
            logger = setup_logger('disassemble', 'disassemble.log')
            for i, infile in tqdm(list(enumerate(infiles))):
                try:
                    print(f'{i} {infile}')
                    disassemble_text(infile, out_path, debug)
                except ValueError as e:
                    errors = True
                    logger.log(ERROR, f'{infile}\n{e}')
        else:
            res = []
            try:
                res += p_uimap(partial(disassemble_text, out_path=out_path), infiles)
            except ValueError as e:
                print(e)
                print('There is the option to run disassembling with `-D` flag which will collect all files with '
                      'mARkdown errors and their error message into a log file.')

    if errors:
        print('Some files had errors and could not be processed, check `disassemble.log`')
    else:
        print('Done')



from glob import glob
from pathlib import Path

import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from functools import partial

from p_tqdm import p_uimap

from eis1600.helper.repo import TRAINING_DATA_REPO
from eis1600.onomastics.methods import nasab_annotation


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-T', '--test', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug
    test = args.test

    if test:
        with open(TRAINING_DATA_REPO + 'gold_standard.txt', 'r', encoding='utf-8') as fh:
            files_txt = fh.read().splitlines()

        infiles = [TRAINING_DATA_REPO + 'gold_standard/' + file for file in files_txt if Path(
                TRAINING_DATA_REPO + 'gold_standard/' + file).exists()]
    else:
        infiles = glob(TRAINING_DATA_REPO + 'training_data_nasab_ML2/*.EIS1600')

    if debug:
        for file in infiles:
            print(file)
            nasab_annotation(file, test)
    else:
        res = []
        res += p_uimap(partial(nasab_annotation, test=test), infiles)

    print('Done')



from datetime import date
from functools import partial
from typing import Dict, Optional
from pathlib import Path
from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from re import compile
from json import dump

from p_tqdm import p_uimap
from numpy import nan

from eis1600.helper.repo import TOPO_TRAINING_REPO, TRAINING_DATA_REPO
from eis1600.markdown.md_to_bio import md_to_bio
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.toponyms.toponym_categories import TOPONYM_CATEGORIES, TOPONYM_LABEL_DICT, TOPONYM_LESS_LABEL_DICT, TOPONYM_CATEGORIES_REPLACEMENTS

CATS = ''.join(TOPONYM_CATEGORIES)
TOP_PATTERN = compile(r"T(?P<num_tokens>\d)(?P<cat>[" + CATS + "])")


def reconstruct_automated_tag(row) -> str:
    return 'ÜT' + row['num_tokens'] + row['cat']


def get_tops_true(file: str, keep_automatic_tags: Optional[bool] = False) -> Dict:
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object, keep_automatic_tags)

    s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
    df_true = s_notna.str.extract(TOP_PATTERN).dropna(how='all')
    tops = df_true.apply(reconstruct_automated_tag, axis=1)
    tops.name = 'TOPS_TRUE'

    if not tops.empty:
        df = df.join(tops)
    else:
        df['TOPS_TRUE'] = nan

    bio_tags = md_to_bio(
            df[['TOKENS', 'TOPS_TRUE']],
            'TOPS_TRUE',
            TOP_PATTERN,
            'TO',
            TOPONYM_LESS_LABEL_DICT,
            True,
            TOPONYM_CATEGORIES_REPLACEMENTS
    )

    return bio_tags


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug
    keep = True

    with open(TRAINING_DATA_REPO + 'gold_standard.txt', 'r', encoding='utf-8') as fh:
        files_txt = fh.read().splitlines()

    infiles = [TRAINING_DATA_REPO + 'gold_standard_topo/' + file for file in files_txt if Path(
            TRAINING_DATA_REPO + 'gold_standard_topo/' + file).exists()]
    
    # with open(TRAINING_DATA_REPO + '5k_gold_standard.txt', 'r', encoding='utf-8') as fh:
    #     files_txt = fh.read().splitlines()

    # infiles = [TRAINING_DATA_REPO + '5k_gold_standard/' + file for file in files_txt if Path(
    #         TRAINING_DATA_REPO + '5k_gold_standard/' + file
    # ).exists()]

    res = []
    if debug:
        for file in infiles[40:60]:
            print(file)
            res.append(get_tops_true(file, keep))
    else:
        res += p_uimap(partial(get_tops_true, keep_automatic_tags=keep), infiles)

    with open(
            TOPO_TRAINING_REPO + 'toponyms_category_training_data_' + date.today().isoformat() + '.json',
            'w',
            encoding='utf-8'
    ) as fh:
        dump(res, fh, indent=4, ensure_ascii=False)

    print('Done')



from re import compile

from openiti.helper.ara import denormalize

from eis1600.helper.ar_normalization import normalize_dict


# TODO Annotate hajja -> Mekka
# TODO Annotate jāwara -> Medina

TOPONYM_CATEGORIES_DICT = {
        'ولد': 'B', 'مولد': 'B',
        'مات': 'D', 'موت': 'D', 'توفي': 'D', 'وفاة': 'D',
        'دفن': 'G',
        'سمع': 'K', 'روى': 'K', 'روا': 'K', 'قرا': 'K', 'اجاز': 'K', 'حدث': 'K',
        'استقر': 'R', 'انفصل': 'R', 'ولي': 'R', 'قاضي': 'R', 'نائب': 'R', 'صاحب': 'R', 'أعمال': 'R',
        # 'حج': 'V',
        'سكن': 'R', 'نزل': 'R', 'نزيل': 'R', 'من اهل': 'R', 'استوطن': 'R', 'كان من': 'R', 'نشأ': 'R'
}

TOPONYM_CATEGORIES_REPLACEMENTS = {
        'O': 'R',
        'A': 'X',
        'V': 'X'
}

TOPONYM_CATEGORIES = list(set(TOPONYM_CATEGORIES_DICT.values())) + ['X']
TOPONYM_CATEGORIES_NOR = normalize_dict(TOPONYM_CATEGORIES_DICT)

TOPONYM_LABEL_DICT = {
        'B-TOB': 0,
        'I-TOB': 1,
        'B-TOD': 2,
        'I-TOD': 3,
        'B-TOG': 4,
        'I-TOG': 5,
        'B-TOK': 6,
        'I-TOK': 7,
        'B-TOR': 8,
        'I-TOR': 9,
        'B-TOX': 10,
        'I-TOX': 11,
        'O': 12
}

TOPONYM_LESS_LABEL_DICT = {
        'B-TOB': 0,
        'I-TOB': 1,
        'B-TOD': 2,
        'I-TOD': 3,
        'B-TOR': 4,
        'I-TOR': 5,
        'B-TOX': 6,
        'I-TOX': 7,
        'O': 8
}

AR_TOPONYM_CATEGORIES = '|'.join([denormalize(key) for key in TOPONYM_CATEGORIES_DICT.keys()])
TOPONYM_CATEGORY_PATTERN = compile(r'\s[وف]?(?P<topo_category>' + AR_TOPONYM_CATEGORIES + r')')



from re import compile

from openiti.helper.ara import denormalize

from eis1600.helper.ar_normalization import normalize_dict
from eis1600.helper.markdown_patterns import WORD

ONES = {
        'واحد': 1, 'احدى': 1, 'احد': 1, 'اثنين': 2, 'اثنتين': 2, 'اثنتي': 2, 'ثلاث': 3, 'ثلث': 3, 'اربع': 4, 'خمس': 5,
        'ست': 6, 'سبع': 7, 'ثماني': 8, 'ثمان': 8, 'تسع': 9, 'نيف': 5, 'بضع': 5
}
ONES_NOR = normalize_dict(ONES)
ONES_HINDI = {
        '٠': 0, '١': 1, '٢': 2, '٣': 3, '٤': 4, '٥': 5, '٦': 6, '٧': 7, '٨': 8, '٩': 9
}
TEN = {
        'عشرة': 10, 'عشري': 10, 'عشر': 10, 'عشرين': 20, 'ثلاثين': 30, 'اربعين': 40, 'خمسين': 50, 'ستين': 60,
        'سبعين': 70,
        'ثمانين': 80, 'تسعين': 90
}
TEN_NOR = normalize_dict(TEN)
HUNDRED = {
        'مائة': 100, 'ماية': 100, 'مية': 100, 'مئة': 100,
        'المائة': 100, 'الماية': 100, 'المية': 100, 'المئة': 100,
        'مائتين': 200, 'مايتين': 200, 'ميتين': 200,
        'ثلاثمائة': 300, 'ثلاث مائة': 300, 'ثلثمائة': 300, 'ثلث مائة': 300, 'ثلاثماية': 300, 'ثلاث ماية': 300,
        'ثلثماية': 300, 'ثلث ماية': 300, 'ثلاثمية': 300, 'ثلاث مية': 300, 'ثلثمية': 300, 'ثلث مية': 300,
        'ثلاثمئة': 300, 'ثلاث مئة': 300, 'ثلثمئة': 300, 'ثلث مئة': 300,
        'اربعمائة': 400, 'اربع مائة': 400, 'اربعماية': 400, 'اربع ماية': 400, 'اربعمية': 400, 'اربع مية': 400,
        'اربعمئة': 400, 'اربع مئة': 400,
        'خمسمائة': 500, 'خمس مائة': 500, 'خمسماية': 500, 'خمس ماية': 500, 'خمسمية': 500, 'خمس مية': 500, 'خمسمئة': 500,
        'خمس مئة': 500,
        'ستمائة': 600, 'ست مائة': 600, 'ستماية': 600, 'ست ماية': 600, 'ستمية': 600, 'ست مية': 600, 'ستمئة': 600,
        'ست مئة': 600,
        'سبعمائة': 700, 'سبع مائة': 700, 'سبعماية': 700, 'سبع ماية': 700, 'سبعمية': 700, 'سبع مية': 700, 'سبعمئة': 700,
        'سبع مئة': 700,
        'ثمانمائة': 800, 'ثمان مائة': 800, 'ثمانيمائة': 800, 'ثماني مائة': 800, 'ثمانماية': 800, 'ثمان ماية': 800,
        'ثمانيماية': 800, 'ثماني ماية': 800, 'ثمانمية': 800, 'ثمان مية': 800, 'ثمانيمية': 800, 'ثماني مية': 800,
        'ثمانمئة': 800, 'ثمان مئة': 800, 'ثمانيمئة': 800, 'ثماني مئة': 800,
        'تسعمائة': 900, 'تسع مائة': 900, 'تسعماية': 900, 'تسع ماية': 900, 'تسعمية': 900, 'تسع مية': 900, 'تسعمئة': 900,
        'تسع مئة': 900
}
HUNDRED_NOR = normalize_dict(HUNDRED)
THOUSAND = normalize_dict({'ألف': 1000})
THOUSAND_NOR = normalize_dict(THOUSAND)

DAY_ONES = {
        'واحد': 1, 'حادي': 1, 'ثاني': 2, 'ثالث': 3, 'رابع': 4, 'خامس': 5, 'خميس': 5, 'سادس': 6, 'سابع': 7,
        'ثامن': 8, 'تاسع': 9, 'عاشر': 10
}
DAY_ONES_NOR = normalize_dict(DAY_ONES)
DAY_TEN = {'عشرة': 10, 'عشري': 10, 'عشر': 10, 'عشرين': 20, 'عشرون': 20, 'ثلاثين': 30, 'ثلاثون': 30}
DAY_TEN_NOR = normalize_dict(DAY_TEN)
WEEKDAYS = {
        'يوم الأحد': 1, 'يوم الاثنين': 2, 'يوم الثلاثاء': 3, 'يوم الأربعاء': 4, 'يمو الخميس': 5, 'يوم الجمعة': 6,
        'يوم السبت': 7
}
WEEKDAYS_NOR = normalize_dict(WEEKDAYS)

MONTHS = {
        'محرم': 1, 'شهر الله المحرم': 1, 'صفر': 2, 'صفر الخير': 2, 'ربيع': 3, 'ربيع الاول': 3, 'ربيع الثاني': 4,
        'ربيع الاخر': 4, 'جمادى الاول': 5, 'جمادى الاولى': 5, 'جمادى الاخرة': 6, 'جمادى الاخر': 6, 'جمادى الثانية': 6,
        'رجب': 7, 'رجب الفرد': 7, 'رجب المبارك': 7, 'شعبان': 8, 'شعبان المكرم': 8, 'رمضان': 9,
        'رمضان المعظم': 9, 'شوال': 10, 'ذي القعدة': 11, 'ذي قعدة': 11, 'ذي الحجة': 12, 'ذي حجة': 12, 'ذو القعدة': 11,
        'ذو قعدة': 11, 'ذو الحجة': 12, 'ذو حجة': 12, 'اخر': -1
}
MONTHS_NOR = normalize_dict(MONTHS)
SANA = ['سنة', 'عام', 'في حدود']

AR_MONTHS = '|'.join([denormalize(key) for key in MONTHS.keys()])
AR_ONES = '|'.join([denormalize(key) for key in ONES.keys()])
AR_TEN = '|'.join([denormalize(key) for key in TEN.keys()])
AR_HUNDRED = '|'.join([denormalize(key) for key in HUNDRED.keys()])
AR_THOUSAND = '|'.join([denormalize(key) for key in THOUSAND.keys()])
AR_ONES_DAY = '|'.join([denormalize(key) for key in DAY_ONES.keys()])
AR_TEN_DAY = '|'.join([denormalize(key) for key in DAY_TEN.keys()])
AR_WEEKDAY = '|'.join([denormalize(key) for key in WEEKDAYS.keys()])
AR_SANA = '|'.join([denormalize(s) for s in SANA])
DATE = r'(?P<context>' + WORD + r'{0,10}?' + r'(?:\s(?:ف[يى]|تقريبا))?' + WORD + r'{0,9}?)' + \
       r'(?:\s(?P<weekday>' + AR_WEEKDAY + r'))?' + \
       r'(?:\s(:?ال)?(?P<day_ones>' + AR_ONES_DAY + r'))?(?:\s(:?و)?(:?ال)?(?P<day_ten>' + AR_TEN_DAY + r'))?' + \
       r'(?:\s(?:(?:من\s)?(?:شهر\s)?)?(?:ال)?(?P<month>' + AR_MONTHS + r')(?:\s(?:من|ف[يى])(?:\sشهور)?)?)?' + \
       r'(?P<year>' + \
              r'\s(?P<sana>' + AR_SANA + ')' + \
              r'(?:\s(?P<digits_str>(?P<digits>[٠١٢٣٤٥٦٧٨٩\d]{1,4})(?:\s(?:[هم]|الهجري[هة]))?))?' + \
              r'(?:\s(?P<ones>' + AR_ONES + r'))?' + \
              r'(?:\s[و]?(?P<ten>' + AR_TEN + r'))?' + \
              r'(?:\s[و]?(?P<hundred>' + AR_HUNDRED + r'))?' + \
              r'(?:\s[و]?(?P<thousand>' + AR_THOUSAND + r'))?' + \
       r')' + \
       r'(?=(?:' + WORD + r'|[\s.,]|$))'

DATE_PATTERN = compile(DATE)
MONTH_PATTERN = compile(AR_MONTHS)

DATE_CATEGORIES_DICT = {
       'ولد': 'B', 'مولد': 'B',
       'مات': 'D', 'موت': 'D', 'توفي': 'D', 'وفاة': 'D',
       'حج': 'P',
       'سمع': 'K', 'روى': 'K', 'روا': 'K', 'قرا': 'K', 'اجاز': 'K', 'حدث': 'K',
       'استقر': 'O', 'انفصل': 'O', 'ولي': 'O', 'قاضي': 'O', 'نائب': 'O', 'صاحب': 'O', 'أعمال': 'O'
}
DATE_CATEGORIES_NOR = normalize_dict(DATE_CATEGORIES_DICT)

AR_DATE_CATEGORIES = '|'.join([denormalize(key) for key in DATE_CATEGORIES_DICT.keys()])
DATE_CATEGORY_PATTERN = compile(r'\s[وف]?(?P<date_category>' + AR_DATE_CATEGORIES + r')')

DATE_CATEGORIES = list(set(DATE_CATEGORIES_DICT.values())) + ["X"]

DATE_LABEL_DICT = {
       'B-YYB': 0,
       'I-YYB': 1,
       'B-YYD': 2,
       'I-YYD': 3,
       'B-YYK': 4,
       'I-YYK': 5,
       'B-YYO': 6,
       'I-YYO': 7,
       'B-YYP': 8,
       'I-YYP': 9,
       'B-YYX': 10,
       'I-YYX': 11,
       'O': 12
}

DATE_CATEGORIES_REPLACEMENTS = {
        'M': 'X',
        'R': 'X'
}



from typing import List
import string

from eis1600.markdown.md_to_bio import bio_to_md
from numpy import nan
from pandas import notna

from eis1600.nlp.cameltools import lemmatize_and_tag_ner, CamelToolsModels


def annotate_miu_text(df):
    lemmas, ner_tags, pos_tags, root_tags, st_tags, fco_tags, toponym_tags = ['_'], ['_'], ['_'], ['_'], ['_'], ['_'], ['_']
    section_id, temp_tokens = None, []
    # TODO STOP processing per section - rather, use overlapping windows of tokens. Sections are not a reliable unit
    for entry in list(zip(df['SECTIONS'].to_list(), df['TOKENS'].fillna('-').to_list()))[1:]:
        _section, _token = entry[0], entry[1]
        if _section is not None:
            # Start a new section
            if len(temp_tokens) > 0:
                # 1. process the previous section
                _labels = lemmatize_and_tag_ner(temp_tokens)
                _, _ner_tags, _lemmas, _dediac_lemmas, _pos_tags, _root_tags, _st_tags, _fco_tags, _toponym_tags = zip(
                        *_labels)
                ner_tags.extend(_ner_tags)
                lemmas.extend(_dediac_lemmas)
                pos_tags.extend(_pos_tags)
                root_tags.extend(_root_tags)
                fco_tags.extend(_fco_tags)
                st_tags.extend(_st_tags)
                toponym_tags.extend(_toponym_tags)
                # 2. reset variables
                section_id, temp_tokens = None, []

        token = _token if _token not in ['', None] else '_'
        temp_tokens.append(token)

    if len(temp_tokens) > 0:
        _labels = lemmatize_and_tag_ner(temp_tokens)
        _, _ner_tags, _lemmas, _dediac_lemmas, _pos_tags, _root_tags, _st_tags, _fco_tags, _toponym_tags = zip(*_labels)
        ner_tags.extend(_ner_tags)
        lemmas.extend(_dediac_lemmas)
        pos_tags.extend(_pos_tags)
        root_tags.extend(_root_tags)
        fco_tags.extend(_fco_tags)
        st_tags.extend(_st_tags)
        toponym_tags.extend(_toponym_tags)

    return ner_tags, lemmas, pos_tags, root_tags, st_tags, fco_tags, toponym_tags


def aggregate_STFCON_classes(st_list: list, fco_list: list) -> List[str]:
    label_dict = {
        "B-FAMILY": "F",
        "I-FAMILY": "F",
        "B-CONTACT": "C",
        "I-CONTACT": "C",
        "B-OPINION": "O",
        "I-OPINION": "O",
        "O": "",
        "_": "",
        "B-TEACHER": "T",
        "I-TEACHER": "T",
        "B-STUDENT": "S",
        "I-STUDENT": "S",
        "I-NEUTRAL": "X",
        "B-NEUTRAL": "X",
    }
    aggregated_labels = []
    for a, b in zip(st_list, fco_list):
        labels = f"{label_dict.get(a, '')}{label_dict.get(b, '')}"
        if a == b:
            labels = label_dict.get(a, '')
        else:
            labels = labels.replace("X", "")
        aggregated_labels.append(labels)
    return aggregated_labels


def merge_ner_with_person_classes(ner_labels, aggregated_stfco_labels):
    merged_labels = []
    for a, b in zip(ner_labels, aggregated_stfco_labels):
        if notna(a) and a[:2] == "ÜP":
            postfix = "X"
            if b.strip() != "":
                postfix = b.strip()
            a = a + postfix
        merged_labels.append(a)
    return merged_labels


def merge_ner_with_toponym_classes(ner_labels: List[str], toponym_labels: List[str]) -> List[str]:
    merged_labels = []
    for a, b in zip(ner_labels, toponym_labels):
        if notna(a) and a[:2] == 'ÜT':
            if b:
                merged_labels.append(b)
            else:
                merged_labels.append(a + 'X')
        else:
            merged_labels.append(a)
    return merged_labels


def insert_onom_tag(df) -> list:
    tokens = df['TOKENS'].fillna('-').to_list()
    nasab_tagger = CamelToolsModels.getNasabModel()
    shortend_list_of_tokens = tokens[1:]
    __shortend_list_limit = 120
    if len(tokens) > __shortend_list_limit:
        shortend_list_of_tokens = tokens[1:__shortend_list_limit]
    for idx, t in enumerate(shortend_list_of_tokens):
        if t.strip() == "":
            shortend_list_of_tokens[idx] = "-"
    nasab_labels = nasab_tagger.predict_sentence(shortend_list_of_tokens)
    punct = "..،_" + string.punctuation
    nasab = ['_']
    nasab_started = False
    for token, label in zip(shortend_list_of_tokens, nasab_labels):
        if label == "B-NASAB":
            # Start a new NASAB
            nasab.append("BONOM")
            nasab_started = True
        elif label == "I-NASAB":
            nasab.append(nan)
        else:
            if nasab_started:
                nasab.append("EONOM")
                nasab_started = False
            else:
                nasab.append(nan)
    if nasab_started:
        nasab[-1] = "EONOM"
    # merge the shortend list
    if len(tokens) > __shortend_list_limit:
        nasab.extend([nan] * (len(tokens) - __shortend_list_limit))
    return nasab


def insert_onomastic_tags(df):
    onomastic_tagger = CamelToolsModels.getOnomasticModel()
    onomastic_tags = [nan] * len(df['TOKENS'])
    start_nasab_id, end_nasab_id = -1, -1

    # Find BNASAB & ENASAB
    for idx, tag in enumerate(df['ONOM_TAGS'].to_list()):
        if "BONOM" == tag:
            start_nasab_id = idx
        elif "EONOM" == tag:
            end_nasab_id = idx
            break

    if 0 < start_nasab_id < end_nasab_id:
        nasab_tokens = df['TOKENS'].fillna('-').to_list()[start_nasab_id:end_nasab_id]
        onomastic_labels = onomastic_tagger.predict_sentence(nasab_tokens)
        ono_tags = bio_to_md(onomastic_labels)

        for i, tag in enumerate(ono_tags):
            onomastic_tags[start_nasab_id + i] = tag

    return onomastic_tags



from glob import glob
from os.path import isfile, splitext
from sys import argv
from argparse import Action, ArgumentParser, RawDescriptionHelpFormatter
from p_tqdm import p_uimap
from re import compile

from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.processing.postprocessing import write_updated_miu_to_file
from eis1600.helper.markdown_patterns import onom_tags
from eis1600.helper.repo import TRAINING_DATA_REPO

PATTERN = compile(r'(ÜP\d+[A-Z]),((?:BONOM,)?Ü(?:' + onom_tags + r')\d)')


def fix_ono_o_tag_order(file) -> bool:
    with open(file, 'r+', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object, keep_automatic_tags=True)
        s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
        s_replaced = s_notna.str.replace(PATTERN, lambda m: m.group(2) + ',' + m.group(1))
        s_new = s_replaced.apply(lambda tags_str: tags_str.split(','))
        df.loc[s_replaced.index, 'TAGS_LISTS'] = s_new

        write_updated_miu_to_file(
                miu_file_object, yml_handler, df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']]
        )

    return True


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.IDs' and fileext != '.EIS1600':
                parser.error('You need to input an IDs file or a single MIU file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to generate JSON from MIU YAMLHeaders.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='IDs or MIU file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    debug = args.debug
    file = args.input

    if file:
        fix_ono_o_tag_order(file)
    else:
        infiles = glob(TRAINING_DATA_REPO + '5k_gold_standard/*.EIS1600')
        res = []

        if debug:
            for i, file in enumerate(infiles[:10]):
                print(f'{i} {file}')
                res.append(fix_ono_o_tag_order(file))
        else:
            res += p_uimap(fix_ono_o_tag_order, infiles)

    print('Done')



from os.path import split, splitext
from typing import List, Tuple

from eis1600.helper.markdown_patterns import CATEGORY_PATTERN, HEADER_END_PATTERN, HEADING_PATTERN, MIU_TAG_PATTERN, \
    MIU_UID_PATTERN, PAGE_TAG_PATTERN
from eis1600.miu.HeadingTracker import HeadingTracker
from eis1600.miu.methods import check_file_for_mal_formatting
from eis1600.miu.yml_handling import create_yml_header


def get_text_as_list_of_mius(infile: str) -> List[Tuple[str, str, bool]]:
    """Disassemble text into list of MIUs.

    Splits the texts into individual MIUs and returns a list of all contained MIUs.
    :param str infile: File which will be disassembled.
    :return List[Tuple[str, str, bool]]: returns the individual MIUs as tuples of (uid, miu_text, analysis_flag).
    """
    heading_tracker = HeadingTracker()
    path, uri = split(infile)
    uri, ext = splitext(uri)
    uid = ''
    miu_text = ''

    with open(infile, 'r', encoding='utf8') as text:
        header_text = text.read().split('#META#Header#End#')

        try:
            check_file_for_mal_formatting(infile, header_text[1])
        except ValueError:
            raise

        mal_formatted = []
        mius = []

        text.seek(0)
        for text_line in iter(text):
            if HEADER_END_PATTERN.match(text_line):
                miu_text = ''
                uid = uri + '.' + 'preface'
                next(text)  # Skip empty line after header
            elif MIU_UID_PATTERN.match(text_line):
                if HEADING_PATTERN.match(text_line):
                    m = HEADING_PATTERN.match(text_line)
                    heading_text = m.group('heading')
                    if PAGE_TAG_PATTERN.search(heading_text):
                        heading_text = PAGE_TAG_PATTERN.sub('', heading_text)
                    heading_tracker.track_headings(len(m.group('level')), heading_text)
                if miu_text:
                    # Do not create a preface MIU file if there is no preface
                    mius.append((uid, miu_text + '\n', analyse_flag))
                m = MIU_TAG_PATTERN.match(text_line)
                uid = uri + '.' + m.group('UID')
                category = ''
                try:
                    category = CATEGORY_PATTERN.search(m.group('category')).group(0)
                except AttributeError:
                    mal_formatted.append(m.group('category'))
                yml_header = create_yml_header(category, heading_tracker.get_curr_state())
                analyse_flag = category not in ['|PARATEXT|', '|EDITOR|']
                miu_text = yml_header
                miu_text += text_line
            else:
                miu_text += text_line

            if PAGE_TAG_PATTERN.search(text_line):
                heading_tracker.track_pages(PAGE_TAG_PATTERN.search(text_line).group(0))

        # last MIU needs to be written to file when the for-loop is finished
        mius.append((uid, miu_text + '\n', analyse_flag))

    if mal_formatted:
        print('Something seems to be mal-formatted, check:')
        print(infile + '\n')
        for elem in mal_formatted:
            print(elem)

    return mius



from __future__ import annotations

from ast import literal_eval
from typing import Any, Dict, Optional

from eis1600.helper.markdown_patterns import MIU_HEADER
from eis1600.helper.yml_methods import dict_to_yaml
from eis1600.miu.HeadingTracker import HeadingTracker


class YAMLHandler:
    """A class to take care of the MIU YAML Headers

    :param Dict yml: the YAML header as a dict, optional.
    :ivar Literal['NOT REVIEWED', 'REVIEWED'] reviewed: Indicates if the file has manually been reviewed, defaults to
    'NOT REVIEWED'.
    :ivar str reviewer: Initials of the reviewer if the file was already manually reviewed, defaults to None.
    :ivar HeadingTracker headings: HeadingTracker returned by the get_curr_state method of the HeaderTracker.
    :ivar List[str] dates_headings: List of dates tags contained in headings.
    :ivar List[int] dates: List of dates contained in the text.
    :ivar Dict onomstics: contains onomastic elements by category.
    :ivar str category: String categorising the type of the entry, bio, chr, dict, etc.
    """
    # Only attributes named in the following list are allowed to be added to the YAMLHeader - add any new attribute
    # to that list
    __attr_from_annotation = ['dates', 'min_date', 'max_date', 'ages', 'onomastics', 'ambiguous_toponyms', 'persons',
                              'toponyms', 'settlements', 'provinces', 'edges_settlements', 'edges_provinces',
                              'books', 'miscs']

    @staticmethod
    def __parse_yml_val(val: str) -> Any:
        if val.isdigit():
            return int(val)
        if len(val.split('.')) == 2 and val.split('.')[0].isdigit() and val.split('.')[1].isdigit():
            return float(val)
        elif val == 'True':
            return True
        elif val == 'False':
            return False
        elif val == 'None' or val == '':
            return None
        elif val.startswith(('\'', '"')):
            return val.strip('\'"')
        elif val.startswith('['):
            # List - no comma allowed in strings, it is used as the separator!
            raw_val_list = val[1:-1]  # strip '[]' but without stripping multiple in case we have nested lists
            if raw_val_list.startswith('(') and raw_val_list.endswith(')'):
                # List of tuples
                val_list = raw_val_list.strip('()').split('), (')
                values = []
                for v in val_list:
                    t = v.split(', ')
                    values.append((YAMLHandler.__parse_yml_val(t[0]), YAMLHandler.__parse_yml_val(t[1])))
            elif raw_val_list.startswith('[') or raw_val_list.startswith('{'):
                # Nested lists
                values = literal_eval(val)
            elif raw_val_list == '':
                return None
            else:
                # List of other values
                val_list = raw_val_list.split(', ')
                values = [YAMLHandler.__parse_yml_val(v) for v in val_list]
            return values
        else:
            return val

    @staticmethod
    def __parse_yml(yml_str: str) -> Dict:
        yml = {}
        level = []
        for line in yml_str.splitlines():
            if not line.startswith('#'):
                intend = (len(line) - len(line.lstrip())) / 4
                key_val = line.split(':')
                key = key_val[0].strip(' -')
                val = ':'.join(key_val[1:]).strip()

                while intend < len(level):
                    # Go as many levels up as necessary, for each level: add key, dict to the parent level and pop child
                    dict_key = level[-1][0]
                    dict_val = level[-1][1]
                    if len(level) > 1:
                        level[-2][1][dict_key] = dict_val
                    else:
                        yml[dict_key] = dict_val
                    level.pop()

                if intend and intend == len(level) and val != '':
                    # Stay on level and add key, val to the respective dict
                    curr_dict = level[-1][1]
                    curr_dict[key] = YAMLHandler.__parse_yml_val(val)
                elif val == '':
                    # Go one level deeper, add key and empty dict for that new level
                    level.append((key, {}))
                else:
                    # Add key, val to the top level
                    yml[key] = YAMLHandler.__parse_yml_val(val)

        while len(level) > 0:
            dict_key = level[-1][0]
            dict_val = level[-1][1]
            if len(level) > 1:
                level[-2][1][dict_key] = dict_val
            else:
                yml[dict_key] = dict_val
            level.pop()

        # This is fix for old files, nas should be part of onomastics
        if hasattr(yml, 'nas'):
            delattr(yml, 'nas')

        return yml

    def __init__(self, yml: Optional[Dict] = None) -> None:
        self.reviewed = 'NOT REVIEWED'
        self.reviewer = 'RESEARCHER'
        self.category = None
        self.headings = None
        self.dates_headings = None
        self.number_of_tokens = None

        for key in YAMLHandler.__attr_from_annotation:
            if key == 'ambiguous_toponyms':
                self.__setattr__(key, False)
            else:
                self.__setattr__(key, None)

        if yml:
            for key, val in yml.items():
                if key == 'headings':
                    val = HeadingTracker(val)
                if key == 'ambigious':
                    # Fix typo
                    key = 'ambiguous'
                self.__setattr__(key, val)

    @classmethod
    def from_yml_str(cls, yml_str: str) -> YAMLHandler:
        """Return instance with attr set from the yml_str."""
        return cls(YAMLHandler.__parse_yml(yml_str))

    def set_category(self, category: str) -> None:
        self.category = category

    def set_number_of_tokens(self, number_of_tokens: int) -> None:
        self.number_of_tokens = number_of_tokens

    def set_ambiguous_toponyms(self) -> None:
        self.ambiguous_toponyms = True

    def set_headings(self, headings: HeadingTracker) -> None:
        self.headings = headings

    def set_reviewed(self, reviewer) -> None:
        self.reviewed = 'REVIEWED'
        self.reviewer = reviewer

    def unset_reviewed(self) -> None:
        self.reviewed = 'NOT REVIEWED'
        self.reviewer = None

    def set_error_while_collecting_annotated_entities(self, tag: str) -> None:
        self.error_while_collecting_annotated_entities = True
        if hasattr(self, 'erroneous_tags'):
            self.erroneous_tags.append(tag)
        else:
            self.erroneous_tags = [tag]

    def get_yamlfied(self) -> str:
        yaml_str = MIU_HEADER + 'Begin#\n\n'
        for key, val in vars(self).items():
            if val:
                if key == 'category':
                    yaml_str += key + '    : \'' + val + '\'\n'
                elif hasattr(val, 'get_yamlfied'):
                    yaml_str += f'{key}    : {val.get_yamlfied()}\n'
                elif isinstance(val, dict):
                    yaml_str += f'{key}    :\n{dict_to_yaml(val, 1)}\n'
                else:
                    yaml_str += key + '    : ' + str(val) + '\n'
        yaml_str += '\n' + MIU_HEADER + 'End#\n\n'

        return yaml_str

    def to_json(self, init) -> Dict:
        json_dict = init
        for key, val in vars(self).items():
            if val:
                if hasattr(val, 'to_json'):
                    json_dict[key] = val.to_json()
                else:
                    json_dict[key] = val
        return json_dict

    def is_bio(self) -> bool:
        return self.category == '$' or self.category == '$$'

    def is_reviewed(self) -> bool:
        return self.reviewed.startswith('REVIEWED')

    def add_date_headings(self, date: int) -> None:
        if self.dates_headings:
            if date not in self.dates_headings:
                self.dates_headings.append(date)
        else:
            self.dates_headings = [date]

    def add_tagged_entities(self, entities_dict: dict) -> None:
        for key in YAMLHandler.__attr_from_annotation:
            # Clear old entities
            if key != 'ambiguous_toponyms':
                self.__setattr__(key, None)
            elif hasattr(self, 'ambiguous_toponyms'):
                self.__delattr__('ambiguous_toponyms')
        for key in YAMLHandler.__attr_from_annotation:
            # Set new entities in same order
            if key in entities_dict.keys():
                self.__setattr__(key, entities_dict.get(key))

    def __setitem__(self, key: str, value: Any) -> None:
        super().__setattr__(key, value)

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __str__(self) -> str:
        return self.get_yamlfied()



from threading import Lock
from typing import List

from camel_tools.ner import NERecognizer


class Model:
    __model = None
    __lock = Lock()

    def __init__(self, model_path: str) -> None:
        Model.__model = NERecognizer(model_path)
        Model.__lock = Lock()

    @staticmethod
    def predict_sentence(tokens: List[str]) -> List[str]:
        with Model.__lock:
            return Model.__model.predict_sentence(tokens)



from typing import Tuple

from eis1600.helper.markdown_patterns import AGE_PATTERN, YEAR_PATTERN


def get_yrs_tag_value(tag: str) -> Tuple[int, str]:
    """Returns int value encoded in the tag for date and age tags.

    :param str tag: Annotation.
    :return int:  numeric value which was encoded in the tag.
    :raise ValueError: If the tag is not correct a ValueError is raised.
    """
    if YEAR_PATTERN.match(tag):
        m = YEAR_PATTERN.match(tag)
    elif AGE_PATTERN.match(tag):
        m = AGE_PATTERN.match(tag)
    else:
        raise ValueError

    if m.group('real'):
        return int(m.group('real')), m.group('cat')
    else:
        return int(m.group('written')), m.group('cat')



import json
from numpy import floating, integer, ndarray


class EvalResultsEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, integer):
            return int(obj)
        if isinstance(obj, floating):
            return float(obj)
        if isinstance(obj, ndarray):
            return obj.tolist()
        return super(EvalResultsEncoder, self).default(obj)



from argparse import ArgumentParser, RawDescriptionHelpFormatter
from glob import glob
from os.path import split
from sys import argv

from p_tqdm import p_uimap
from tqdm import tqdm

from eis1600.helper.repo import MIU_REPO, TOPO_REPO
from eis1600.markdown.md_to_bio import bio_to_md
from eis1600.models.ToponymDescriptionModel import ToponymDescriptionModel
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.processing.postprocessing import merge_tagslists, reconstruct_miu_text_with_tags


def annotate_miu(file: str) -> str:
    outpath = TOPO_REPO + 'data/' + split(file)[1]
    
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

    toponym_labels = ToponymDescriptionModel().predict_sentence(df['TOKENS'].fillna('-').to_list())
    if 'B-TOPD' in toponym_labels:
        df['Q'] = bio_to_md(toponym_labels, umlaut_prefix=False)
        df['TAGS_LISTS'] = df.apply(lambda x: merge_tagslists(x['TAGS_LISTS'], x['Q']), axis=1)

        yml_handler.unset_reviewed()
        updated_text = reconstruct_miu_text_with_tags(df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']])

        with open(outpath, 'w', encoding='utf-8') as ofh:
            ofh.write(str(yml_handler) + updated_text)

    return outpath


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate toponym descriptions in MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    infiles = glob(MIU_REPO + 'data/*/*/MIUs/*[0-9].EIS1600')

    res = []
    if debug:
        for i, file in tqdm(list(enumerate(infiles))):
            print(i, file)
            res.append(annotate_miu(file))
    else:
        res += p_uimap(annotate_miu, infiles)

    print('Done')



from sys import argv, exit
from os.path import isfile, split, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from pathlib import Path

from numpy import number
from p_tqdm import p_uimap
from pandas import DataFrame

from eis1600.helper.repo import RESEARCH_DATA_REPO, get_files_from_eis1600_dir, read_files_from_readme, MIU_REPO
from eis1600.miu.methods import get_mius
from eis1600.statistics.methods import count_tokens


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.IDs' and fileext != '.EIS1600':
                parser.error('You need to input an IDs file or a single MIU file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to count tokens per MIU file(s).
-----
Give an IDs file or a single MIU file as input
otherwise 
all files in the MIU directory are batch processed.
'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='IDs or MIU file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    debug = args.debug

    if args.input:
        infile = './' + args.input
        filepath, fileext = splitext(infile)
        if fileext == '.IDs':
            mius = get_mius(infile)[1:]  # First element is path to the OPENITI HEADER
            print(f'NER annotate MIUs of {infile}')
            res = []
            if debug:
                for i, miu in enumerate(mius):
                    print(f'{i} {miu}')
                    res.append(count_tokens(miu))
            else:
                res += p_uimap(count_tokens, mius)
        else:
            print(f'NER annotate {infile}')
            count_tokens(infile)
    else:
        input_dir = MIU_REPO

        if not Path(input_dir).exists():
            print('Your working directory seems to be wrong, make sure it is set to the parent dir of '
                  '`EIS1600_MIUs/`.')
            exit()

        data_path = RESEARCH_DATA_REPO + 'token_count/data/'
        Path(data_path).mkdir(parents=True, exist_ok=True)
        stats_path = RESEARCH_DATA_REPO + 'token_count/stats/'
        Path(stats_path).mkdir(parents=True, exist_ok=True)


        print(f'Count tokens per MIU')
        files_list = read_files_from_readme(input_dir, '# Texts disassembled into MIU files\n')
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'IDs')
        if not infiles:
            print('There are no IDs files to process')
            exit()

        for n, infile in enumerate(infiles):
            file_path, uri = split(infile)
            uri, ext = splitext(uri)
            if not debug:
                print(f'{n} {uri}')

            mius = get_mius(infile)[1:]  # First element is path to the OPENITI HEADER
            res = []
            if debug:
                for i, miu in enumerate(mius):
                    print(f'{i} {miu}')
                    res.append(count_tokens(miu))
            else:
                res += p_uimap(count_tokens, mius)

            df = DataFrame(res, columns=['URI', 'CATEGORY', 'NUMBER_OF_TOKENS'])
            df = df.astype({'CATEGORY': 'category', 'NUMBER_OF_TOKENS': 'int32'})
            df.to_csv(data_path + uri + '.csv', index=False)
            
            df.describe(include=[number, 'category']).to_csv(stats_path + uri + '_info.csv')
            df.groupby('CATEGORY').describe().to_csv(stats_path + uri + '_info_per_type.csv')

    print('Done')



from functools import partial

from tqdm import tqdm
from p_tqdm import p_uimap
from pathlib import Path

from sys import argv, exit
from os.path import isfile, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from glob import glob

from eis1600.helper.repo import get_files_from_eis1600_dir, read_files_from_readme, write_to_readme, \
    update_texts_fixed_poetry_readme
from eis1600.markdown.methods import convert_to_EIS1600TMP


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext not in ['.mARkdown', '.inProgess', '.completed']:
                parser.error('You need to input a mARkdown file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to convert mARkdown file(s) to EIS1600TMP file(s).
-----
Give a single mARkdown file as input
or 
Give an input AND an output directory for batch processing.

Run without input arg to batch process all mARkdown files in the EIS1600 directory which have not been processed yet.
'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='MARkdown file to process or input directory with mARkdown files to process if an output directory is '
                 'also given',
            action=CheckFileEndingAction
    )
    arg_parser.add_argument(
            'output', type=str, nargs='?',
            help='Optional, if given batch processes all files from the input directory to the output directory'
    )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input and not args.output:
        infile = './' + args.input
        if 'data' in infile:
            path = infile.split('data')[0]
        else:
            depth = len(infile.split('/'))
            if depth == 2:
                path = '../../../'
            elif depth == 3:
                path = '../../'
            else:
                path = '../'
        print(f'Convert mARkdown file {infile} to EIS1600TMP file')
        convert_to_EIS1600TMP(infile, None, verbose)
        infiles = [infile.split('/')[-1]]
        write_to_readme(path, infiles, '# Texts converted into `.EIS1600TMP`\n', '.EIS1600TMP')

    elif args.output:
        input_dir = args.input
        output_dir = args.output
        if not input_dir[-1] == '/':
            input_dir += '/'

        print(f'Convert mARkdown files from {input_dir}, save resulting EIS1600TMP files to {output_dir}')

        infiles = glob(input_dir + '*.mARkdown')
        if not infiles:
            print(
                    'The input directory does not contain any mARkdown files to process'
            )
            exit()

        # Check if output directory exists else create that directory
        Path(output_dir).mkdir(exist_ok=True, parents=True)

        if verbose:
            for infile in tqdm(infiles):
                try:
                    convert_to_EIS1600TMP(infile, output_dir, verbose)
                except Exception as e:
                    print(infile, e)
        else:
            res = []
            res += p_uimap(partial(convert_to_EIS1600TMP, output_dir=output_dir), infiles)

    else:
        input_dir = './'

        print(f'Update list of texts with automatically fixed poetry in the README')
        update_texts_fixed_poetry_readme(input_dir, '# Texts with fixed poetry\n')
        print(f'List of texts with automatically fixed poetry was successfully updated in the README')

        print(f'Convert mARkdown files from the EIS1600 repo (only if there is not an EIS1600TMP file yet)')
        files_list = read_files_from_readme(input_dir, '# Texts with fixed poetry\n')
        infiles = get_files_from_eis1600_dir(
                input_dir, files_list, ['mARkdown', 'inProcress', 'completed'], 'EIS1600*'
        )
        if not infiles:
            print(
                    'There are no more mARkdown files to process'
            )
            exit()

        if verbose:
            for infile in tqdm(infiles):
                try:
                    convert_to_EIS1600TMP(infile, None, verbose)
                except Exception as e:
                    print(infile, e)
        else:
            res = []
            res += p_uimap(convert_to_EIS1600TMP, infiles)

        write_to_readme(input_dir, infiles, '# Texts converted into `.EIS1600TMP`\n', '.EIS1600TMP')

    print('Done')



from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
from sys import argv, exit
from typing import Optional
from logging import ERROR
from time import process_time, time

import jsonpickle
from tqdm import tqdm
from p_tqdm import p_uimap

from torch import cuda

from eis1600.corpus_analysis.miu_methods import analyse_miu
from eis1600.corpus_analysis.text_methods import get_text_as_list_of_mius
from eis1600.helper.logging import setup_logger
from eis1600.helper.repo import JSON_REPO, TEXT_REPO, get_files_from_eis1600_dir, read_files_from_autoreport


def routine_per_text(infile: str, debug: Optional[bool] = False):
    """Entry into analysis routine per text.

    Each text is disassembled into the list of MIUs. Analysis is applied to each MIU. Writes a JSON file containing
    the list of MIUs with their analysis results.
    :param ste infile: EIS1600 text which is analysed.
    :param bool debug: Debug flag for serial processing, otherwise parallel processing.
    """
    mius_list = get_text_as_list_of_mius(infile)

    res = []
    if debug:
        for idx, tup in tqdm(list(enumerate(mius_list[:20]))):
            res.append(analyse_miu(tup))
    else:
        res += p_uimap(analyse_miu, mius_list[:20])

    out_path = infile.replace(TEXT_REPO, JSON_REPO)
    out_path = out_path.replace('.EIS1600', '.json')
    dir_path = '/'.join(out_path.split('/')[:-1])
    Path(dir_path).mkdir(parents=True, exist_ok=True)

    with open(out_path, 'w', encoding='utf-8') as fh:
        jsonpickle.set_encoder_options('json', indent=4, ensure_ascii=False)
        json_str = jsonpickle.encode(res, unpicklable=False)
        fh.write(json_str)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to parse whole corpus to annotated MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    print(f'GPU available: {cuda.is_available()}')

    st = time()
    stp = process_time()

    # Retrieve all double-checked texts
    input_dir = TEXT_REPO
    files_list = read_files_from_autoreport(input_dir)

    infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600')
    if not infiles:
        print('There are no EIS1600 files to process')
        exit()

    logger = setup_logger('disassemble', 'disassemble.log')
    for i, infile in tqdm(list(enumerate(infiles[:5]))):
        try:
            print(f'{i} {infile}')
            routine_per_text(infile, debug)
        except ValueError as e:
            errors = True
            logger.log(ERROR, f'{infile}\n{e}')

    et = time()
    etp = process_time()

    print('Done')
    print(f'Processing time: {etp-stp} seconds')
    print(f'Execution time: {et-st} seconds')



from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
from sys import argv
from glob import iglob
from re import compile
from typing import List, Optional, Tuple

from p_tqdm import p_uimap
from pandas import concat, DataFrame, read_csv
from openiti.helper.ara import denormalize

from eis1600.gazetteers.Spellings import Spellings
from eis1600.gazetteers.Toponyms import Toponyms
from eis1600.helper.markdown_patterns import WORD, NOISE_ELEMENTS
from eis1600.helper.repo import MIU_REPO, TOPO_REPO
from eis1600.processing.preprocessing import get_yml_and_miu_df

place_terms = ['كورة', 'كور', 'قرية', 'قرى', 'مدينة', 'مدن', 'ناحية', 'نواح', 'نواحي', 'محلة', 'محال', 'محلات', 'بلد',
               'بلاد', 'ارباع', 'رستاق', 'رساتيق', 'أعمال']  # 'ربع'
technical_terms = ['من', 'بين',
                   'نسبة',
                   'يوم', 'يوما',
                   'مرحلة', 'مرحلتان', 'مرحلتين', 'مراحل',
                   'فرسخ', 'فرسخا', 'فراسخ',
                   'ميل', 'ميلا', 'أميال']
dn_pt = [denormalize(t) for t in place_terms]
dn_tt = [denormalize(t) for t in technical_terms]
dn_spelling = Spellings().get_denormalized_list()
dn_toponyms = Toponyms().total()


PLACES_REGEX = compile(
    r'(?P<context>(?:' + WORD + NOISE_ELEMENTS +
    r'(?P<spelling> [وب]?(?:ال)?(?:' + '|'.join(dn_spelling) + ')(?:ها)?' + NOISE_ELEMENTS + r')*){1,10})'
    r' (?P<place_term>' + '|'.join(dn_pt) + r')(?P<more_context>' + WORD + NOISE_ELEMENTS + r'){1,7}'
    )
TT_REGEX = compile(r'|'.join(dn_pt + dn_tt + dn_spelling + dn_toponyms))


def annotate_miu(file: str, debug: Optional[bool] = False) -> List[Tuple[str, str, str]]:
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

    miu = Path(file).name.replace('.EIS1600', '')
    passages = []

    text = ' '.join(df['TOKENS'].loc[df['TOKENS'].notna()].to_list())
    text_updated = text

    if debug:
        print(text)

    if PLACES_REGEX.search(text_updated):
        m = PLACES_REGEX.search(text_updated)
        while m:
            start = m.start()
            end = m.end()
            if debug:
                print(m.group(0))
                print('Context :', m.group('context'))
                print('Place Term: ', m.group('place_term'))
            if len(TT_REGEX.findall(m.group(0))) >= 3:
                passages.append((miu, m.group(0), m.group(0)))
                text_updated = text_updated[:start] + ' BTOPD' + text_updated[start:end] + ' ETOPD' + text_updated[end:]
                m = PLACES_REGEX.search(text_updated, end + 12)
            else:
                m = PLACES_REGEX.search(text_updated, end)

    return passages


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 file to process'
    )

    args = arg_parser.parse_args()
    debug = args.debug

    if args.input:
        author, text, version, uid = args.input.split('.')
        file_path = MIU_REPO + 'data/' + '/'.join([author, '.'.join([author, text])]) + \
                    '/MIUs/' + args.input + '.EIS1600'
        annotate_miu(file_path, debug)
    else:
        sheets = iglob(TOPO_REPO + 'sheet_[0-9].csv')
        sheets_df = DataFrame(None, columns=['MIU', 'ORIGINAL', 'MODIFIABLE', 'STATUS'])

        for sheet in sheets:
            tmp = read_csv(sheet).dropna()
            sheets_df = concat([sheets_df, tmp])

        sheets_df['STATUS'].loc[sheets_df['STATUS'].str.fullmatch('TOPOPONYM')] = 'TOPONYM'
        # entries = sheets_df.loc[sheets_df['STATUS'].str.fullmatch('TOPONYM|NISBA|CORRECT')]
        incomplete_mius = sheets_df.loc[sheets_df['STATUS'].str.fullmatch('INCOMPLETE'), 'MIU'].to_list()

        infiles = []

        for miu in incomplete_mius:
            author, text, version, uid = miu.split('.')
            file_path = MIU_REPO + 'data/' + '/'.join([author, '.'.join([author, text])]) + \
                        '/MIUs/' + miu + '.EIS1600'
            infiles.append(file_path)

        res = []
        if debug:
            for i, file in enumerate(infiles[:20]):
                print(i, file)
                res.append(annotate_miu(file))
        else:
            res += p_uimap(annotate_miu, infiles)

        tuples = []
        [tuples.extend(r) for r in res if r]

        df = DataFrame(tuples, columns=['MIU', 'ORIGINAL', 'MODIFIABLE'])
        df['STATUS'] = None
        df.to_csv(TOPO_REPO + 'topod_incomplete.csv', index=False)

    print('Done')



from re import compile
from glob import glob
from p_tqdm import p_uimap
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.processing.postprocessing import write_updated_miu_to_file


sheikhuna = compile('[و]?شيخنا')


def untag_sheikhuna(row):
    if row['TOKENS'] and sheikhuna.match(row['TOKENS']) and row['TAGS_LISTS'] and [tag for tag in row['TAGS_LISTS']
                                                                                   if tag.startswith('P1')]:
        if len(row['TAGS_LISTS']) > 1:
            row['TAGS_LISTS'].pop()
            return row['TAGS_LISTS'], True
        else:
            return None, True
    else:
        return row['TAGS_LISTS'], False


def check_file(file):
    with open(file, 'r+', encoding='utf-8') as fh:
        fh.readline()
        fh.readline()
        if fh.readline() == 'reviewed    : REVIEWED2\n':
            fh.seek(0)
            yml, df = get_yml_and_miu_df(fh)
            df['TAGS_LISTS'], changed = zip(*df[['TOKENS', 'TAGS_LISTS']].apply(untag_sheikhuna, axis=1))
            if any(changed):
                write_updated_miu_to_file(fh, yml, df)


def main():
    files = glob('*.EIS1600')

    res = []
    res += p_uimap(check_file, files)






from tqdm import tqdm
from p_tqdm import p_uimap
from pathlib import Path

from sys import argv, exit
from os.path import isfile, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from glob import glob
from multiprocessing import Pool

from eis1600.helper.repo import get_files_from_eis1600_dir, write_to_readme, read_files_from_readme
from eis1600.markdown.methods import insert_uids


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.EIS1600TMP':
                parser.error('You need to input an EIS1600TMP file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to insert UIDs in EIS1600TMP file(s) and thereby converting them to final EIS1600 
            file(s).
-----
Give a single EIS1600TMP file as input
or 
Give an input AND an output directory for batch processing.

Run without input arg to batch process all EIS1600TMP files in the EIS1600 directory which have not been processed yet.
'''
            )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600TMP file to process, you need to run this command from inside text repo',
            action=CheckFileEndingAction
            )
    arg_parser.add_argument(
            'output', type=str, nargs='?',
            help='Optional, if given batch processes all files from the input directory to the output directory'
            )
    args = arg_parser.parse_args()

    verbose = args.verbose

    # If this script is run with a single file as input
    if args.input and not args.output:
        infile = './' + args.input
        if 'data' in infile:
            path = infile.split('data')[0]
        else:
            depth = len(infile.split('/'))
            if depth == 2:
                path = '../../../'
            elif depth == 3:
                path = '../../'
            else:
                path = '../'
        print(f'Insert UIDs into {infile}')
        insert_uids(infile, None, verbose)
        infiles = [infile.split('/')[-1]]
        write_to_readme(path, infiles, '# Texts converted into `.EIS1600`\n', '.EIS1600', True)

    # If this script is run with input and output directory (old, not used any more)
    elif args.output:
        input_dir = args.input
        output_dir = args.output

        print(f'Insert UIDs into {input_dir}, save resulting EIS1600 files to {output_dir}')

        infiles = glob(input_dir + '/*.EIS1600TMP')
        if not infiles:
            print(
                    'The input directory does not contain any EIS1600TMP files to process'
                    )
            exit()

        # Check if output directory exists else create that directory
        Path(output_dir).mkdir(exist_ok=True, parents=True)

        params = [(infile, output_dir, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(insert_uids, params).get()

    # If this script is run with -e option on the EIS1600 text repo
    else:
        input_dir = './'

        # Get EIS1600TMP files which are marked as ready to process in the README and have not yet been converted to
        # EIS1600
        print(f'Insert UIDs into files from the EIS1600 repo (only if there is not an EIS1600 file yet)')
        files_list = read_files_from_readme(input_dir, '# Texts converted into `.EIS1600TMP`\n')
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600TMP', 'EIS1600')
        if not infiles:
            print(
                    'There are no more EIS1600TMP files to process'
                    )
            exit()

        # Insert UIDs in the selected files, runs in parallel
        if verbose:
            for infile in tqdm(infiles):
                try:
                    insert_uids(infile, None, verbose)
                except Exception as e:
                    print(infile, e)
        else:
            res = []
            res += p_uimap(insert_uids, infiles)

        # Write list of processed files to the README
        write_to_readme(input_dir, infiles, '# Texts converted into `.EIS1600`\n', '.EIS1600', True)

    print('Done')



from typing import List, Literal, Optional, Union
from pandas import Series
from openiti.helper.ara import normalize_ara_heavy

from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.processing.postprocessing import write_updated_miu_to_file
from eis1600.toponyms.toponym_categories import TOPONYM_CATEGORIES_NOR, TOPONYM_CATEGORY_PATTERN


def add_category_to_tag(
        tag_or_tag_list: Union[str, List[str]],
        category: Literal['B', 'D', 'K', 'M', 'O', 'P', 'R', 'X'],
) -> Union[str, List[str]]:
    if isinstance(tag_or_tag_list, list):
        return [t if not t.startswith('T') else 'Ü' + t + category for t in tag_or_tag_list]
    else:
        return tag_or_tag_list + category


def toponym_category_annotate_miu(s_tokens: Series, s_tags: Series) -> Series:
    """Categorize toponyms based on their context and add category to their tag.

    :param Series s_tokens: Series of tokens, needed for context analysis of the toponym.
    :param Series s_tags: Series of NER tags or TAGS_LISTS, containing the toponyms tags which shall be altered.
    :return Series: Series of altered tags.
    """

    s_notna = None
    s_new_tags = s_tags.copy()
    if s_tags.name == 'TAGS_LISTS':
        s_notna = s_tags.loc[s_tags.notna()].apply(lambda tag_list: ','.join(tag_list))
    else:
        s_notna = s_tags.loc[s_tags.notna()]

    if s_notna.empty:
        return s_tags

    toponym_idcs = s_notna.loc[s_notna.str.contains(r'T\d')].index

    for idx in toponym_idcs:
        # Get context
        min_idx = idx - 10 if idx - 10 > 0 else 0
        tokens = s_tokens.loc[min_idx:idx]
        context = ' '.join([t for t in tokens if isinstance(t, str)])

        if TOPONYM_CATEGORY_PATTERN.search(context):
            # Assign category based on last match in context
            last = TOPONYM_CATEGORY_PATTERN.findall(context)[-1]
            toponym_category = TOPONYM_CATEGORIES_NOR.get(normalize_ara_heavy(last))
        else:
            toponym_category = 'X'

        # Update tag
        s_new_tags.loc[idx] = add_category_to_tag(s_tags.loc[idx], category=toponym_category)

    return s_new_tags


def toponym_category_annotation(file: str, test: Optional[bool] = False, keep_automatic_tags: Optional[bool] = False):
    """Helper to run toponyms category annotation standalone as cmdline script.

    :param str file: Path of the miu file to annotate.
    :param bool test: Optional, indicating if the script is run with test data, defaults to false.
    :param bool keep_automatic_tags: Optional, if True keep automatic annotation (Ü-tags), defaults to False.
    """
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object, keep_automatic_tags)

    df['TAGS_LISTS'] = toponym_category_annotate_miu(df['TOKENS'], df['TAGS_LISTS'])

    if test:
        output_path = str(file).replace('gold_standard_nasab', 'gold_standard_topo')
    else:
        output_path = str(file)

    with open(output_path, 'w', encoding='utf-8') as out_file_object:
        write_updated_miu_to_file(
                out_file_object, yml_handler, df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']]
        )



from collections import Counter
from functools import partial

import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from p_tqdm import p_uimap

from eis1600.stats.methods import extract_categories
from eis1600.helper.repo import read_files_from_readme, get_files_from_eis1600_dir


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to generate statistics on MIU files.'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    args = arg_parser.parse_args()

    verbose = args.verbose

    input_dir = './'

    print(f'Statistics on MIUs')
    files_list = read_files_from_readme(input_dir, '# Texts disassembled into MIU files\n')

    infiles = get_files_from_eis1600_dir(input_dir, files_list, 'YAMLDATA.yml')
    if not infiles:
        print('There are no YAMLDATA files to process')
        sys.exit()

    counter_categories = Counter()
    counter_reviewers = Counter()

    res = []
    res += p_uimap(partial(extract_categories, verbose=verbose), infiles)
    categories = [cat for cats, revs in res for cat in cats]
    reviewers = [rev for cats, revs in res for rev in revs]
    counter_categories.update(categories)
    counter_reviewers.update(reviewers)

    with open('STATS.md', 'w', encoding='utf-8') as stats_fh:
        stats_fh.write('# Stats\n\n')
        stats_fh.write('## Categories\n\n')

        for key, val in counter_categories.items():
            line = key + '  : ' + str(val) + '\n'
            stats_fh.write(line)
        stats_fh.write('\ntotal    : ' + str(sum(counter_categories.values())) + '\n\n')

        stats_fh.write('## Reviewers\n\n')

        for key, val in counter_reviewers.items():
            line = key + '  : ' + str(val) + '\n'
            stats_fh.write(line)
        stats_fh.write('\ntotal    : ' + str(sum(counter_reviewers.values())) + '\n\n')

    print('Done')



from argparse import Action, ArgumentParser, RawDescriptionHelpFormatter
from glob import glob
from os.path import isdir, isfile
from sys import argv
from time import process_time, time

from p_tqdm import p_uimap

from eis1600.miu.methods import annotate_miu_file
from torch import cuda
from tqdm import tqdm


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and (isdir(input_arg) or isfile(input_arg)):
            setattr(namespace, self.dest, input_arg)
        else:
            print('You need to specify a valid path to the directory holding the files which shall be (re-)annotated')
            raise IOError


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotated directory holding gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='Directory which holds the files to process or individual file to annotate',
            action=CheckFileEndingAction
    )

    args = arg_parser.parse_args()
    debug = args.debug
    input_df = args.input

    print(f'GPU available: {cuda.is_available()}')

    if isfile(input_df):
        annotate_miu_file(input_df)
    else:
        st = time()
        stp = process_time()

        mius = glob(input_df + '*.EIS1600')

        if debug:
            for idx, miu in tqdm(list(enumerate(mius))):
                try:
                    annotate_miu_file(miu)
                except Exception as e:
                    print(idx, miu)
                    print(e)
        else:
            res = []
            res += p_uimap(annotate_miu_file, mius)

        et = time()
        etp = process_time()

        print(f'Processing time: {etp-stp} seconds')
        print(f'Execution time: {et-st} seconds')

    print('Done')



from os.path import isdir

from camel_tools.ner import NERecognizer
from camel_tools.tokenizers.word import simple_word_tokenize
from camel_tools.disambig.mle import MLEDisambiguator
from camel_tools.tagger.default import DefaultTagger
from camel_tools.utils.dediac import dediac_ar
from typing import Iterator, Tuple, Union

from eis1600.helper.repo import PRETRAINED_MODELS_REPO


class CamelToolsModels:
    __pos_tagger = None
    __mled_disambiguator = None
    __lemmatizer = None
    __ner_tagger = None
    __nasab_tagger = None
    __onomastic_tagger = None
    __st_tagger = None
    __fco_tagger = None
    __toponym_tagger = None

    __NASAB_MODEL_PATH = PRETRAINED_MODELS_REPO + "camelbert-ca-finetuned_nasab/"
    __NER_MODEL_PATH = PRETRAINED_MODELS_REPO + "camelbert-ca-finetuned_ner/"
    __ONOMASTIC_MODEL_PATH = PRETRAINED_MODELS_REPO + "camelbert-ca-finetuned_onomastic/"
    __STN_MODEL_PATH = PRETRAINED_MODELS_REPO + "camelbert-ca-finetuned_person_classification_STN/"
    __FCON_MODEL_PATH = PRETRAINED_MODELS_REPO + "camelbert-ca-finetuned_person_classification_FCN/"
    __TOPO_MODEL_PATH = PRETRAINED_MODELS_REPO + "camelbert-ca-finetuned_toponyms/"

    @staticmethod
    def getInstance():
        """ Static access method. """
        if CamelToolsModels.__ner_tagger is None or CamelToolsModels.__lemmatizer is None:
            CamelToolsModels()
        return CamelToolsModels.__mled_disambiguator, CamelToolsModels.__ner_tagger, CamelToolsModels.__lemmatizer, \
               CamelToolsModels.__pos_tagger, CamelToolsModels.__st_tagger, CamelToolsModels.__fco_tagger, \
               CamelToolsModels.__toponym_tagger

    @staticmethod
    def getNasabModel():
        """ Static access method. """
        if CamelToolsModels.__nasab_tagger is None:
            CamelToolsModels.__nasab_tagger = NERecognizer(CamelToolsModels.__NASAB_MODEL_PATH)
        return CamelToolsModels.__nasab_tagger

    @staticmethod
    def getOnomasticModel():
        """ Static access method. """
        if CamelToolsModels.__onomastic_tagger is None:
            CamelToolsModels.__onomastic_tagger = NERecognizer(CamelToolsModels.__ONOMASTIC_MODEL_PATH)
        return CamelToolsModels.__onomastic_tagger

    def __init__(self):
        """ Virtually private constructor. """
        if not isdir(PRETRAINED_MODELS_REPO):
            raise Exception(f'We miss our custom ML models in {PRETRAINED_MODELS_REPO}')

        if CamelToolsModels.__ner_tagger is not None:
            raise Exception("This class is a singleton!")
        else:
            CamelToolsModels.__mled_disambiguator = MLEDisambiguator.pretrained()
            CamelToolsModels.__lemmatizer = DefaultTagger(CamelToolsModels.__mled_disambiguator, 'lex')
            CamelToolsModels.__pos_tagger = DefaultTagger(CamelToolsModels.__mled_disambiguator, 'pos')
            CamelToolsModels.__ner_tagger = NERecognizer(CamelToolsModels.__NER_MODEL_PATH)  # .pretrained
            CamelToolsModels.__st_tagger = NERecognizer(CamelToolsModels.__STN_MODEL_PATH)
            CamelToolsModels.__fco_tagger = NERecognizer(CamelToolsModels.__FCON_MODEL_PATH)
            CamelToolsModels.__toponym_tagger = NERecognizer(CamelToolsModels.__TOPO_MODEL_PATH)


def lemmatize_and_tag_ner(tokens: Union[str, list]) -> Iterator[Tuple[str, ...]]:
    """Lemmatize the text and annotate named-entities.

    Lemmatize the text and annotated named-entities using Camel Tools models.
    :param tokens: a  string or a list of tokens to be annotated
    """
    mled_disambiguator, ner_tagger, lemmatizer, pos_tagger, st_tager, fco_tagger, toponym_tagger = \
        CamelToolsModels.getInstance()
    # if tokens is a string, then tokenize it
    if isinstance(tokens, str):
        tokens = simple_word_tokenize(tokens)
    ner_labels = ner_tagger.predict_sentence(tokens)
    st_labels = st_tager.predict_sentence(tokens)
    fco_labels = fco_tagger.predict_sentence(tokens)
    toponym_labels = toponym_tagger.predict_sentence(tokens)
    lemmas = lemmatizer.tag(tokens)
    pos_tags = pos_tagger.tag(tokens)
    root_tags = [d.analyses[0].analysis['root'] for d in mled_disambiguator.disambiguate(tokens)]
    dediac_lemmas = [dediac_ar(lemma) for lemma in lemmas]

    return zip(tokens, ner_labels, lemmas, dediac_lemmas, pos_tags, root_tags, st_labels, fco_labels, toponym_labels)



from time import time
from requests import get

from pandas import concat, DataFrame, read_csv

from eis1600.helper.repo import TOPO_REPO


def main():
    sheets = {
        '1': '1vjTQICH_dsJosnhA46Pqbs1ow9CHkptELIiTXDfQS04',
        '2': '1fg56kEs16-_DGdZK2VeR-qtF39U6-qSEQ1p6B5g4zTI',
        '3': '1hThigHwxj1n4mro68eAA4O3wpF2-PuONBmea1x5Yc0o',
        '4': '1lbNzqjdPnonshNU8nVwVTTE7CqCNNhmTpj9CrIXNmYs',
        '5': '12GsdOvkGXWad1v5LUQ9kAyt8ablurFy9wOylcz6SmCY',
        '6': '13cu3OcAMN25VU2X49NYKMoJyPG2CbWFS0LobKSSAwYY',
        '7': '18yBjszyZTZ29c-7Wj7VThPQkJudSxYXSEzJHrZLZxhw',
        '8': '1vxhEZu06BPjbGuljpRXVj60kjsuanvvHfLjVw5Y5tQA',
        '9': '1LtAkJIdujcbEp1qKA31PrpNSIwRryIAuP95BKx-qmNQ',
        'incomplete': '1MPCFunhszG9GNigrGpfebbPE7qdmxQIe6VRF3quIWpE'
    }

    for key, uri in sheets.items():
        start_s = time()
        url = 'https://docs.google.com/spreadsheets/d/' + uri + '/export?format=csv'

        print('Sheet_' + key)
        print(url)

        response = get(url)
        end_s = time()
        print(f'download time: {end_s-start_s}')

        if response.status_code == 200:
            filepath = TOPO_REPO + 'sheet_' + key + '.csv'
            with open(filepath, 'wb') as f:
                f.write(response.content)
        else:
            print(f'Error downloading Google Sheet: {response.status_code}')
            print(f'Sheet_{key} could not be downloaded')

    df = DataFrame(None, columns=['MIU', 'ORIGINAL', 'MODIFIABLE', 'STATUS'])
    for key in sheets.keys():
        tmp = read_csv(TOPO_REPO + 'sheet_' + key + '.csv').dropna()
        df = concat([df, tmp])

    print(df.value_counts(['STATUS']))

    print('Done')
            


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='eis1600',
      version='1.1.6',
      description='EIS1600 project tools and utilities',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/EIS1600/eis1600-pkg',
      author='Lisa Mischer',
      author_email='mischer.lisa@gmail.com',
      license='MIT License',
      packages=find_packages(include=['eis1600', 'eis1600.*'], exclude=['excluded']),
      package_data={'eis1600.gazetteers.data': ['*.csv'], 'eis1600.helper.data': ['*.csv']},
      entry_points={
          'console_scripts': [
                  'analyse_all_on_cluster = eis1600.corpus_analysis.analyse_all_on_cluster:main [EIS]',
                  'annotate_goldstandard = eis1600.nlp.annotate_goldstandard:main [NER]',
                  'annotate_mius = eis1600.nlp.ner_annotate_mius:main [NER]',
                  'annotate_topd = eis1600.helper.annotate_topd:main [NER]',
                  'btopd_to_bio = eis1600.helper.btopd_to_bio:main',
                  'convert_mARkdown_to_EIS1600TMP = eis1600.markdown.convert_mARkdown_to_EIS1600TMP:main',
                  'count_tokens_per_miu = eis1600.statistics.count_tokens_per_miu:main',
                  'disassemble_into_miu_files = eis1600.miu.disassemble_into_miu_files:main',
                  'eval_date_model = eis1600.helper.eval_date_model:main [EVAL]',
                  'eval_topo_cat_model = eis1600.helper.eval_topo_cat_model:main [EVAL]',
                  'fix_ono_p_tags_order = eis1600.helper.fix_ono_p_tags_order:main',
                  'fix_miu_annotation = eis1600.helper.fix_miu_annotation:main',
                  'generate_autoreport = eis1600.helper.generate_autoreport:main',
                  'insert_uids = eis1600.markdown.insert_uids:main',
                  'miu_random_revisions = eis1600.helper.miu_random_revisions:main',
                  'miu_stats = eis1600.stats.miu_stats:main',
                  'onomastic_annotation = eis1600.onomastics.annotation:main',
                  'q_tags_to_bio = eis1600.helper.q_tags_to_bio:main',
                  'reassemble_from_miu_files = eis1600.miu.reassemble_from_miu_files:main',
                  'sheets_topod_stats = eis1600.helper.sheets_topod_stats:main',
                  'topo_tags_to_bio = eis1600.helper.topo_tags_to_bio:main',
                  'topod_extract_incomplete = eis1600.helper.topod_extract_incomplete:main',
                  'topod_extract_places_regex = eis1600.helper.topod_extract_places_regex:main',
                  'topod_insert_into_miu = eis1600.helper.topod_insert_into_miu:main',
                  'toponym_annotation = eis1600.toponyms.annotation:main',
                  'update_uids = eis1600.markdown.update_uids:main',
                  'xx_update_uids_old_process = eis1600.markdown.update_uids_old_process:main',
                  'yml_to_json = eis1600.helper.yml_to_json:main'
          ],
      },
      python_requires='>=3.7',
      install_requires=[
              'openiti',
              'pandas',
              'numpy',
              'tqdm',
              'p_tqdm',
              'importlib_resources',
              'jsonpickle',
              'requests'
      ],
      extras_require={
              'NER': ['camel-tools', 'torch'],
              'EVAL': ['evaluate', 'seqeval', 'tensorflow'],
              'EIS': ['camel-tools', 'torch', 'torchvision', 'torchaudio']
      },
      classifiers=['Programming Language :: Python :: 3',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Development Status :: 1 - Planning',
                   'Intended Audience :: Science/Research']
      )



from os.path import split, splitext
from typing import Tuple

from eis1600.processing.postprocessing import write_updated_miu_to_file

from eis1600.processing.preprocessing import get_yml_and_miu_df


def count_tokens(infile: str) -> Tuple[str, str, int]:
    """Count the number of tokens in this MI.

    :param str infile: Path of the MIU to count the tokens of.
    :return Tuple[str, str, int]: Tuple of the URI, the MIU's type and the number of tokens.
    """
    file_path, uri = split(infile)
    uri, ext = splitext(uri)

    with open(infile, 'r+', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

        number_of_tokens = df['TOKENS'].count()  # Count() counts all non-na values
        category = yml_handler.category

        yml_handler.set_number_of_tokens(number_of_tokens)

        write_updated_miu_to_file(miu_file_object, yml_handler, df)

    return uri, category, number_of_tokens



from typing import Optional

from os.path import split, splitext

from eis1600.markdown.UIDs import UIDs
from eis1600.helper.markdown_patterns import EMPTY_FIRST_PARAGRAPH_PATTERN, EMPTY_PARAGRAPH_PATTERN, \
    HEADER_END_PATTERN, \
    MISSING_DIRECTION_TAG_PATTERN, NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN, MIU_TAG_AND_TEXT_PATTERN, \
    NORMALIZE_BIO_CHR_MD_PATTERN, ONLY_PAGE_TAG_PATTERN, \
    PAGE_TAG_IN_BETWEEN_PATTERN, \
    PAGE_TAG_PATTERN, \
    PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN, SPACES_CROWD_PATTERN, \
    NEWLINES_CROWD_PATTERN, \
    POETRY_PATTERN, SPACES_AFTER_NEWLINES_PATTERN, PAGE_TAG_ON_NEWLINE_MARKDOWN_PATTERN, \
    PAGE_TAG_ON_NEWLINE_TMP_PATTERN, \
    UID_TAG_AND_TEXT_SAME_LINE_PATTERN, \
    UID_PATTERN, \
    HEADING_OR_BIO_PATTERN, \
    BIO_CHR_TO_NEWLINE_PATTERN


def normalize_bio_chr_md(paragraph: str) -> str:
    md = NORMALIZE_BIO_CHR_MD_PATTERN.match(paragraph).group(0)
    if md == '$BIO_MAN$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# $', paragraph)
    elif md == '$BIO_WOM$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# $$', paragraph)
    elif md == '$BIO_REF$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# $$$', paragraph)
    elif md == '$CHR_EVE$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# @', paragraph)
    elif md == '$CHR_RAW$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# @@@', paragraph)
    elif md == '@ RAW':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# @@@', paragraph)
    else:
        return paragraph


def convert_to_EIS1600TMP(infile: str, output_dir: Optional[str] = None, verbose: bool = False) -> None:
    """Coverts a file to EIS1600TMP for review process.

    Converts mARkdown, inProgress, completed file to light EIS1600TMP for the review process. Creates the file with the
    '.EIS1600TMP' extension.

    :param str infile: Path of the file to convert.
    :param str or None output_dir: Directory to write new file to (discontinued), optional.
    :param bool verbose: If True outputs a notification of the file which is currently processed, defaults to False.
    :return None:
    """
    if output_dir:
        path, uri = split(infile)
        uri, ext = splitext(uri)
        outfile = output_dir + '/' + uri + '.EIS1600TMP'
    else:
        path, ext = splitext(infile)
        outfile = path + '.EIS1600TMP'
        path, uri = split(infile)

    if verbose:
        print(f'Convert {uri} from mARkdown to EIS1600 file')

    with open(infile, 'r', encoding='utf8') as infile_h:
        text = infile_h.read()

    header_and_text = HEADER_END_PATTERN.split(text)
    header = header_and_text[0] + header_and_text[1]
    text = header_and_text[2][1:]  # Ignore second new line after #META#Header#End#

    if text[0:2] == '#\n':
        # Some texts start with a plain #, remove these
        text = text[2:]

    # fix
    text = text.replace('~\n', '\n')
    text = text.replace('\n~~', ' ')
    text = text.replace(' \n', '\n')

    # spaces
    text = SPACES_AFTER_NEWLINES_PATTERN.sub('\n', text)
    text = SPACES_CROWD_PATTERN.sub(' ', text)

    # fix poetry
    text = POETRY_PATTERN.sub(r'\1', text)

    # fix page tag on newlines
    text = PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN.sub(r'\1 \2 \3', text)
    text = PAGE_TAG_ON_NEWLINE_MARKDOWN_PATTERN.sub(r' \1\n', text)
    text = SPACES_CROWD_PATTERN.sub(' ', text)

    # fix new lines
    text = text.replace('\n###', '\n\n###')
    text = text.replace('\n# ', '\n\n')
    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)

    text = text.split('\n\n')

    text_updated = []

    for paragraph in text:
        if paragraph.startswith('### '):
            paragraph = paragraph.replace('###', '#')
            if NORMALIZE_BIO_CHR_MD_PATTERN.match(paragraph):
                paragraph = normalize_bio_chr_md(paragraph)
            paragraph = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', paragraph)
        text_updated.append(paragraph)

    # reassemble text
    text = '\n\n'.join(text_updated)
    final = header + '\n\n' + text
    if final[-1] != '\n':
        final += '\n'

    with open(outfile, 'w', encoding='utf8') as outfile_h:
        outfile_h.write(final)


def insert_uids(infile: str, output_dir: Optional[str] = None, verbose: Optional[bool] = False) -> None:
    """Insert UIDs and EIS1600 tags into EIS1600TMP file and thereby convert it to EIS1600 format.


    :param str infile: Path of the file to convert.
    :param str or None output_dir: Directory to write new file to (discontinued), optional.
    :param bool verbose: If True outputs a notification of the file which is currently processed, defaults to False.
    :return None:
    """

    if output_dir:
        path, uri = split(infile)
        uri, ext = splitext(uri)
        outfile = output_dir + '/' + uri + '.EIS1600'
    else:
        path, ext = splitext(infile)
        outfile = path + '.EIS1600'
        path, uri = split(infile)

    if verbose:
        print(f'Insert UIDs into {uri} and convert to final EIS1600 file')

    with open(infile, 'r', encoding='utf8') as infile_h:
        text = infile_h.read()

    # disassemble text into paragraphs
    header_and_text = HEADER_END_PATTERN.split(text)
    header = header_and_text[0] + header_and_text[1]
    text = header_and_text[2].lstrip('\n')  # Ignore new lines after #META#Header#End#
    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
    text = NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.sub('\n\n', text)
    text = text.replace(' \n', '\n')
    text = text.replace('\n ', '\n')
    text = PAGE_TAG_ON_NEWLINE_TMP_PATTERN.sub(r' \1', text)
    text = text.split('\n\n')
    text_updated = []

    uids = UIDs()

    text_iter = text.__iter__()
    paragraph = next(text_iter)
    prev_p = ''

    # Insert UIDs and EIS1600 tags into the text
    while paragraph is not None:
        next_p = next(text_iter, None)

        if paragraph:
            # Only do this if paragraph is not empty
            if paragraph.startswith('#'):
                # Move content to an individual line
                paragraph = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', paragraph)
                paragraph = paragraph.replace('#', f'_ء_#={uids.get_uid()}=')
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if len(heading_and_text) == 2:
                    paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n_ء_ ' + \
                                heading_and_text[1]
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            infile + '\n'
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                text_updated.append(paragraph)
            # TODO elif paragraph.startswith('::')
            elif '%~%' in paragraph:
                paragraph = f'_ء_={uids.get_uid()}= ::POETRY:: ~\n_ء_ ' + '\n_ء_ '.join(paragraph.splitlines())
                text_updated.append(paragraph)
            elif PAGE_TAG_PATTERN.fullmatch(paragraph):
                page_tag = PAGE_TAG_PATTERN.match(paragraph).group('page_tag')
                if PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN.search('\n\n'.join([prev_p, paragraph, next_p])):
                    if text_updated:
                        if text_updated[-1][-1] == ' ':
                            text_updated[-1] += page_tag + ' ' + next_p
                        else:
                            text_updated[-1] += ' ' + page_tag + ' ' + next_p
                        paragraph = next_p
                        next_p = next(text_iter, None)
                elif text_updated:
                    text_updated[-1] += ' ' + page_tag
                # else:
                # Remove PageV00P000 at the beginning in an individual paragraph
                # text_updated.append(paragraph)
                # pass
            else:
                paragraph = f'_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n_ء_ ' + paragraph
                text_updated.append(paragraph)

        prev_p = paragraph
        paragraph = next_p

    # reassemble text
    text = '\n\n'.join(text_updated)
    final = header + '\n\n' + text
    if final[-1] != '\n':
        final += '\n'

    with open(outfile, 'w', encoding='utf8') as outfile_h:
        outfile_h.write(final)


def update_uids(infile: str, verbose: Optional[bool] = False) -> None:
    """Updates a text with missing UIDs and EIS1600 tags.

    :param str infile: Path of the file to update UIDs in.
    :param bool verbose: If True outputs a notification of the file which is currently processed, defaults to False.
    :return None:
    """

    path, ext = splitext(infile)
    outfile = path + '.EIS1600'
    path, uri = split(infile)

    if verbose:
        print(f'Update UIDs in {uri}')

    with open(infile, 'r', encoding='utf8') as infile_h:
        text = infile_h.read()

    # disassemble text into paragraphs
    header_and_text = HEADER_END_PATTERN.split(text)
    header = header_and_text[0] + header_and_text[1]
    text = header_and_text[2].lstrip('\n')  # Ignore new lines after #META#Header#End#
    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
    text = NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.sub('\n\n', text)
    text = MISSING_DIRECTION_TAG_PATTERN.sub('\g<1>_ء_ \g<2>', text)
    text = text.split('\n\n')
    text_updated = []

    used_ids = []

    # Collect existing UIDs
    for idx, paragraph in enumerate(text):
        if UID_PATTERN.match(paragraph):
            uid = int(UID_PATTERN.match(paragraph).group('UID'))
            if uid not in used_ids:
                used_ids.append(uid)
            else:
                # If in the review process an UID was accidentally inserted twice, just remove the second occurrence
                # - this element will get a new UID in the next steps.
                if UID_PATTERN.match(paragraph).group(1):
                    # Python returns None for empty capturing groups which messes up the string
                    text[idx] = UID_PATTERN.sub('\1', paragraph)
                else:
                    text[idx] = UID_PATTERN.sub('', paragraph)

    uids = UIDs(used_ids)

    text_iter = text.__iter__()
    paragraph = next(text_iter)
    prev_p = None

    if EMPTY_FIRST_PARAGRAPH_PATTERN.fullmatch(paragraph):
        # Some OpenITI texts start with a single # which causes a plain UID tag as first element - ignore those
        paragraph = next(text_iter)

    # Insert missing UIDs and EIS1600 tags into the text
    while paragraph is not None:
        next_p = next(text_iter, None)

        if paragraph:
            # Only do this if paragraph is not empty
            if HEADING_OR_BIO_PATTERN.match(paragraph):
                # Move content to an individual line
                paragraph = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', paragraph)
                paragraph = paragraph.replace('#', f'_ء_#={uids.get_uid()}=')
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if len(heading_and_text) == 2:
                    paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n\n_ء_ ' \
                                                      f'{heading_and_text[1]}'
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            infile + '\n'
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
            elif not UID_PATTERN.match(paragraph):
                cat = 'POETRY' if '%' in paragraph else 'UNDEFINED'
                if paragraph.startswith('::'):
                    p_pieces = paragraph.splitlines()
                    section_header = p_pieces[0]
                    if '%' in paragraph:
                        paragraph = '\n_ء_ '.join(p_pieces[1:])
                    elif len(p_pieces) > 2:
                        raise ValueError(
                                infile + '\n'
                                f'There is a single new line in this paragraph:\n{paragraph}'
                        )
                    else:
                        paragraph = p_pieces[1]

                    if paragraph[0] != '_':
                        paragraph = '_ء_ ' + paragraph
                else:
                    if paragraph[0] != '_':
                        paragraph = '_ء_ ' + paragraph
                    section_header = f'::{cat}:: ~'
                paragraph = f'_ء_={uids.get_uid()}= {section_header}\n' + paragraph
            elif UID_TAG_AND_TEXT_SAME_LINE_PATTERN.match(paragraph):
                paragraph = UID_TAG_AND_TEXT_SAME_LINE_PATTERN.sub(r'\1\2\n\3', paragraph)
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if '%' in '\n'.join(heading_and_text[1:]):
                    paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::POETRY:: ~\n\n_ء_' + '\n_ء_ '.join(
                            heading_and_text[1:]
                    )
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            infile + '\n'
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                else:
                    paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n_ء_ ' + \
                                heading_and_text[1]
            elif MIU_TAG_AND_TEXT_PATTERN.match(paragraph):
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if '%' in '\n'.join(heading_and_text[1:]):
                    paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::POETRY:: ~\n_ء_ ' + '\n_ء_ '.join(
                            heading_and_text[1:]
                    )
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            infile + '\n'
                                     f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                else:
                    paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n_ء_ ' + \
                                heading_and_text[1]
            if ONLY_PAGE_TAG_PATTERN.fullmatch(paragraph):
                # Add page tags to previous paragraph if there is no other information contained in the current
                # paragraph
                page_tag = ONLY_PAGE_TAG_PATTERN.match(paragraph).group('page_tag')
                if PAGE_TAG_IN_BETWEEN_PATTERN.search('\n\n'.join([prev_p, paragraph, next_p])):
                    if text_updated:
                        next_p_text = next_p.split('::UNDEFINED:: ~\n_ء_ ')[1]
                        if text_updated[-1][-1] == ' ':
                            text_updated[-1] += page_tag + ' ' + next_p_text
                        else:
                            text_updated[-1] += ' ' + page_tag + ' ' + next_p_text
                        paragraph = next_p
                        next_p = next(text_iter, None)
                elif text_updated:
                    text_updated[-1] += ' ' + page_tag
                # else:
                # Remove PageV00P000 at the beginning in an individual paragraph
                # text_updated.append(paragraph)
                # pass
            elif not EMPTY_PARAGRAPH_PATTERN.fullmatch(paragraph):
                # Do not add empty paragraphs to the updated text
                text_updated.append(paragraph)
        prev_p = paragraph
        paragraph = next_p

    # reassemble text
    text = '\n\n'.join(text_updated)
    final = header + '\n\n' + text
    if final[-1] != '\n':
        final += '\n'

    with open(outfile, 'w', encoding='utf8') as outfile_h:
        outfile_h.write(final)



from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
from sys import argv
from glob import glob
from re import compile
from typing import List, Tuple
from math import floor

from pandas import DataFrame
from p_tqdm import p_uimap
from openiti.helper.ara import denormalize

from eis1600.gazetteers.Spellings import Spellings
from eis1600.gazetteers.Toponyms import Toponyms
from eis1600.helper.markdown_patterns import WORD
from eis1600.helper.repo import MIU_REPO, TOPO_REPO
from eis1600.processing.preprocessing import get_tokens_and_tags, get_yml_and_miu_df
from eis1600.processing.postprocessing import reconstruct_miu_text_with_tags

place_terms = ['كورة', 'كور', 'قرية', 'قرى', 'مدينة', 'مدن', 'ناحية', 'نواح', 'نواحي', 'محلة', 'محال', 'محلات', 'بلد',
              'بلاد', 'ارباع', 'رستاق', 'رساتيق', 'أعمال']      #  'ربع'
technical_terms = ['من', 'بين',
                   'نسبة',
                   'يوم', 'يوما',
                   'مرحلة', 'مرحلتان', 'مرحلتين', 'مراحل',
                   'فرسخ', 'فرسخا', 'فراسخ',
                   'ميل', 'ميلا', 'أميال']
dn_pt = [denormalize(t) for t in place_terms]
dn_tt = [denormalize(t) for t in technical_terms]
dn_spelling = Spellings().get_denormalized_list()
dn_toponyms = Toponyms().total()

PLACES_REGEX = compile(r'(?:' + WORD + '(?: [،.():])?){1,7} (?:' + '|'.join(dn_pt) + r')(?:' + WORD + '){1,7}')
TT_REGEX = compile(r'|'.join(dn_pt + dn_tt + dn_spelling + dn_toponyms))


def annotate_miu(file: str) -> List[Tuple[str, str, str]]:
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

    miu = Path(file).name.replace('.EIS1600', '')
    write_out = False
    passages = []

    text = ' '.join(df['TOKENS'].loc[df['TOKENS'].notna()].to_list())
    text_updated = text

    if PLACES_REGEX.search(text_updated):
        m = PLACES_REGEX.search(text_updated)
        while m:
            start = m.start()
            end = m.end()
            if len(TT_REGEX.findall(m.group(0))) >= 3:
                write_out = True
                passages.append((miu, m.group(0), m.group(0)))
                text_updated = text_updated[:start] + ' BTOPD' + text_updated[start:end] + ' ETOPD' + text_updated[end:]
                m = PLACES_REGEX.search(text_updated, end + 12)
            else:
                m = PLACES_REGEX.search(text_updated, end)

        # if write_out:
        #     ar_tokens, tags = get_tokens_and_tags(text_updated)
        #     df.loc[df['TOKENS'].notna(), 'TAGS_LISTS'] = [[t] if t else t for t in tags]

        #     yml_handler.unset_reviewed()
        #     updated_text = reconstruct_miu_text_with_tags(df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']])

        #     outpath = TOPO_REPO + 'data/' + miu + '.EIS1600'
        #     with open(outpath, 'w', encoding='utf-8') as ofh:
        #         ofh.write(str(yml_handler) + updated_text)

    return passages


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    infiles = glob(MIU_REPO + 'data/*/*/MIUs/*[0-9].EIS1600')

    res = []
    if debug:
        for i, file in enumerate(infiles):
            print(i, file)
            res.append(annotate_miu(file))
    else:
        res += p_uimap(annotate_miu, infiles)

    tuples = []
    [tuples.extend(r) for r in res if r]

    df = DataFrame(tuples, columns=['MIU', 'ORIGINAL', 'MODIFIABLE'])
    df.to_csv(TOPO_REPO + 'topod.csv', index=False)
    c = floor(len(df)/5000)
    i = 0
    while i < c:
        idx = str(i+1)
        while len(idx) < 4:
            idx = str(0) + idx
        df.iloc[i*5000:(i+1)*5000].to_csv(TOPO_REPO + 'topod_' + str(i+1) + '.csv', index=False)
        i += 1
    df.iloc[i*5000:].to_csv(TOPO_REPO + 'topod_' + str(i+1) + '.csv', index=False)

    print('Done')



from __future__ import annotations

from eis1600.helper.Singleton import singleton
from importlib_resources import files
from typing import List, Tuple
from pandas import isna, read_csv

from eis1600.helper.ar_normalization import denormalize_list

toponyms_path = files('eis1600.gazetteers.data').joinpath('toponyms_gazetteer.csv')


@singleton
class Toponyms:
    """
    Gazetteer

    :ivar DataFrame __df: The dataFrame.
    :ivar __settlements List[str]: List of all settlement names and their prefixed variants.
    :ivar __provinces List[str]: List of all province names and their prefixed variants.
    :ivar __total List[str]: List of all toponyms and their prefixed variants.
    :ivar __rpl List[Tuple[str, str]]: List of tuples: expression and its replacement.
    """
    __df = None
    __settlements = None
    __provinces = None
    __total = None
    __rpl = None

    def __init__(self) -> None:
        df = read_csv(toponyms_path, usecols=['URI_GRAVITON', 'LABEL', 'TOPONYM', 'METAREGION', 'TYPE'])
        prefixes = ['ب', 'و', 'وب', 'ل', 'ول']

        def get_all_variations(top: str) -> List[str]:
            variations = denormalize_list(top)
            prefixed_variations = [prefix + t for prefix in prefixes for t in variations]
            return variations + prefixed_variations

        df['TOPONYMS'] = df['TOPONYM'].apply(get_all_variations)

        topos = df.explode('TOPONYMS', ignore_index=True)

        Toponyms.__settlements = topos.loc[topos['TYPE'] != 'province', 'TOPONYMS'].to_list()
        Toponyms.__provinces = topos.loc[topos['TYPE'] == 'province', 'TOPONYMS'].to_list()
        Toponyms.__df = topos
        Toponyms.__df.mask(isna(Toponyms.__df), '', inplace=True)

        Toponyms.__total = Toponyms.__settlements + Toponyms.__provinces
        Toponyms.__rpl = [(elem, elem.replace(' ', '_')) for elem in Toponyms.__total if ' ' in elem]

    @staticmethod
    def settlements() -> List[str]:
        return Toponyms.__settlements

    @staticmethod
    def provinces() -> List[str]:
        return Toponyms.__provinces

    @staticmethod
    def total() -> List[str]:
        return Toponyms.__total

    @staticmethod
    def replacements() -> List[Tuple[str, str]]:
        return Toponyms.__rpl

    @staticmethod
    def look_up_entity(entity: str) -> Tuple[str, str, List[str], List[str]]:
        """Look up tagged entity in settlements (there is no gazetteer of provinces so far).

        :param str entity: The token(s) which were tagged as toponym.
        :return: placeLabel(s) as str, URI(s)-tag as str, list of settlement uri(s), list of province uri(s),
        list of settlement(s) coordinates, list of province(s) coordinates.
        """
        if entity in Toponyms.__total:
            matches = Toponyms.__df.loc[Toponyms.__df['TOPONYMS'].str.fullmatch(entity), ['URI_GRAVITON',
                                                                                          'METAREGION',
                                                                                          'LABEL']]
            uris = matches['URI_GRAVITON'].to_list()
            provinces = [m for m in matches['METAREGION'].to_list() if m != '']
            place = matches['LABEL'].unique()

            return '::'.join(place), '@' + '@'.join(uris) + '@', uris, provinces
        else:
            return entity, '', [], []
